<?php
    /*
     * $Id: maia_db.php 973 2006-03-30 01:46:06Z dmorton $
     *
     * MAIA MAILGUARD LICENSE v.1.0
     *
     * Copyright 2004 by Robert LeBlanc <rjl@renaissoft.com>
     *                   David Morton   <mortonda@dgrmm.net>
     * All rights reserved.
     *
     * PREAMBLE
     *
     * This License is designed for users of Maia Mailguard
     * ("the Software") who wish to support the Maia Mailguard project by
     * leaving "Maia Mailguard" branding information in the HTML output
     * of the pages generated by the Software, and providing links back
     * to the Maia Mailguard home page.  Users who wish to remove this
     * branding information should contact the copyright owner to obtain
     * a Rebranding License.
     *
     * DEFINITION OF TERMS
     *
     * The "Software" refers to Maia Mailguard, including all of the
     * associated PHP, Perl, and SQL scripts, documentation files, graphic
     * icons and logo images.
     *
     * GRANT OF LICENSE
     *
     * Redistribution and use in source and binary forms, with or without
     * modification, are permitted provided that the following conditions
     * are met:
     *
     * 1. Redistributions of source code must retain the above copyright
     *    notice, this list of conditions and the following disclaimer.
     *
     * 2. Redistributions in binary form must reproduce the above copyright
     *    notice, this list of conditions and the following disclaimer in the
     *    documentation and/or other materials provided with the distribution.
     *
     * 3. The end-user documentation included with the redistribution, if
     *    any, must include the following acknowledgment:
     *
     *    "This product includes software developed by Robert LeBlanc
     *    <rjl@renaissoft.com>."
     *
     *    Alternately, this acknowledgment may appear in the software itself,
     *    if and wherever such third-party acknowledgments normally appear.
     *
     * 4. At least one of the following branding conventions must be used:
     *
     *    a. The Maia Mailguard logo appears in the page-top banner of
     *       all HTML output pages in an unmodified form, and links
     *       directly to the Maia Mailguard home page; or
     *
     *    b. The "Powered by Maia Mailguard" graphic appears in the HTML
     *       output of all gateway pages that lead to this software,
     *       linking directly to the Maia Mailguard home page; or
     *
     *    c. A separate Rebranding License is obtained from the copyright
     *       owner, exempting the Licensee from 4(a) and 4(b), subject to
     *       the additional conditions laid out in that license document.
     *
     * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER AND CONTRIBUTORS
     * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
     * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
     * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
     * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
     * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
     * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
     * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
     * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
     * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
     * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
     *
     */

    require_once ("core.php");
    require_once ("DB.php");     // Pear::DB
    require_once ("mailtools.php");
    require_once ("smtp.php");
    if (extension_loaded('mcrypt')) {
        require_once ("encrypt.php");
    }


    if (! isset($maia_sql_connect_options)) {
       $maia_sql_connect_options = array();
    }

    // Establish a database connection.
    $dbh = DB::connect($maia_sql_dsn, $maia_sql_connect_options);
    if (DB::isError($dbh)) {
        die($lang['error_connect'] . $dbh->getMessage());
    }
    $dbh->setFetchMode(DB_FETCHMODE_ASSOC);

    /*
     * who_am_i(): Returns a string describing the effective user, and the
     *             administrator (if the admin is impersonating a user).
     */
    function who_am_i($uid, $euid)
    {
       global $lang;

       $fake_user = get_user_name($euid);
       if ($uid != $euid) {
       	  $real_user = get_user_name($uid);
       	  if (is_a_domain_default_user($euid)) {
       	      if (is_system_default_user($euid)) {
       	          $i_am = $lang['text_banner_system_user'];
       	      } else {
       	          $i_am = sprintf($lang['text_banner_domain_user'], $fake_user);
       	      }
       	  } else {
       	      $i_am = sprintf($lang['text_banner_administrator'], $real_user, $fake_user);
       	  }
       } else {
          $i_am = sprintf($lang['text_banner_user'], $fake_user);
       }
       return ($i_am);
    }


    /*
     * ok_to_impersonate(): Returns true if user $uid is allowed to
     *                      impersonate user $euid, false otherwise.
     *                      Note that this is only used to determine
     *                      whether the impersonator should be allowed
     *                      to view and administer the user's mail
     *                      items.  This test is not necessary for
     *                      adjusting a user's settings.
     */
    function ok_to_impersonate($euid, $uid)
    {
        global $dbh;
        
    	// It's harmless to impersonate yourself ;)
        if (($euid == $uid) && ($euid > 0) && ($uid > 0)) {
            return true;
        } else {

    	    // Domain default users can be impersonated by admins
    	    // responsible for those domains, and the superadmin.
    	    // Only the superadmin can impersonate the system default
    	    // user (@.).
            if (is_a_domain_default_user($euid) || get_config_value("enable_privacy_invasion") == "Y") {

              if (is_superadmin($uid)) {
                  return true;
              } else {
                if (is_a_domain_default_user($euid)) {
                  $domain_id = get_domain_id(get_user_name($euid));
                    return (is_admin_for_domain($uid, $domain_id));
                } else if (!is_superadmin($euid)) {
      
                    $select = "SELECT email FROM users WHERE maia_user_id = ?";
                    $sth = $dbh->query($select, array($euid));
                    while ($row = $sth->fetchRow()) {
                      $domain_id = get_domain_id("@" . get_domain_from_email($row["email"]));
                      if (is_admin_for_domain($uid, $domain_id)) {
                          $sth->free();
                          return true;
                      }
                    }
                    $sth->free();
                    return false;
  
                } else {
                    return false;
                }     
              }
              // Impersonating other users is an invasion of privacy,
              // even for administrators, unless explicitly overridden above.
            } else {
                return false;
            }
        }
    }


    /*
     * get_user_name(): Returns the name of the specified user.  For
     *                  auto-created accounts, this will be the same
     *                  as the user's e-mail address, until the user
     *                  finally logs in.
     */
    function get_user_name($uid)
    {
        global $dbh;

        $user_name = "Unknown";
        $select = "SELECT user_name FROM maia_users WHERE id = ?";
        $sth = $dbh->query($select, array($uid));
        if ($row = $sth->fetchRow()) {
            $user_name = $row["user_name"];
        }
        $sth->free();

        return $user_name;
    }


    /*
     * get_user_id(): Looks up the Maia ID of the specified user
     *                by name. Renames database entry from $email
     *                to $user_name if $user_name is not found
     *                and $email is found.
     */
    function get_user_id($user_name, $email)
    {
        global $dbh;

        $select = "SELECT id FROM maia_users " .
                  "WHERE user_name = ?";
        $sth = $dbh->query($select, array($user_name));

        // Try to look the user up by name first.
        if ($row = $sth->fetchRow()) {

            $uid = $row["id"];
            $sth->free();

        } else {

            $sth->free();

            // Now try looking the user up by e-mail address,
            // in case the user was "auto-created".
            $sth = $dbh->query($select, array($email));
            if ($row = $sth->fetchRow()) {
               $uid = $row["id"];
               $sth->free();

               // Set the user's proper name in the database
               // record, to speed up future lookups.
               $update = "UPDATE maia_users SET user_name = ? " .
                         "WHERE id = ?";
               $dbh->query($update, array($user_name, $uid));

            } else {

               // Not found!
               $uid = 0;

            }
        }

        return $uid;
    }


    /*
     * get_user_level(): Returns the user's privilege level,
     *                   one of:
     *
     *                   'U' : User
     *                   'A' : Domain Administrator
     *                   'S' : Superadministrator
     */
    function get_user_level($uid)
    {
        global $dbh;

        $select = "SELECT user_level FROM maia_users " .
                  "WHERE id = ?";
        $sth = $dbh->query($select, array($uid));
        $user_level = "U";
        if ($row = $sth->fetchRow()) {
            $user_level = $row["user_level"];
        }
        $sth->free();

    	return $user_level;
    }


    /*
     * get_domain_id(): Looks up a domain name (in @domain format)
     *                  and returns its id if found, 0 otherwise.
     */
    function get_domain_id($domain_name)
    {
        global $dbh;

        $domain_id = 0;
        $select = "SELECT id FROM maia_domains WHERE domain = ?";
        $sth = $dbh->query($select, array(strtolower($domain_name)));
        if ($row = $sth->fetchrow()) {
            $domain_id = $row["id"];
        }
        $sth->free();

        return $domain_id;
    }


    /*
     * is_admin_for_domain(): Returns true if the specified user
     *                        has administrator privileges for
     *                        the specified domain, false otherwise.
     */
    function is_admin_for_domain($uid, $domain_id)
    {
        global $dbh;

        $user_level = get_user_level($uid);

        if ($user_level == "S") {
            return true;
        } else {
            $select = "SELECT maia_users.user_level " .
                      "FROM maia_users, maia_domain_admins " .
                      "WHERE maia_users.id = ? " .
                      "AND maia_users.id = maia_domain_admins.admin_id " .
                      "AND maia_domain_admins.domain_id = ?";

            $sth = $dbh->query($select, array($uid, $domain_id));
            if ($row = $sth->fetchrow()) {
                $result = ($row["user_level"] == "A");
            } else {
                $result = false;
            }
            $sth->free();

            return $result;
        }
    }


    /*
     * is_an_administrator(): Returns true if the specified user is
     *                        the superadministrator or the admin
     *                        of one or more domains, false otherwise.
     */
    function is_an_administrator($uid)
    {
    	$user_level = get_user_level($uid);

    	return ($user_level == "A" || $user_level == "S");
    }


    /*
     * is_superadmin(): Returns true if the specified user is the
     *                  superadministrator, false otherwise.
     */
    function is_superadmin($uid)
    {
    	$user_level = get_user_level($uid);

    	return ($user_level == "S");
    }


    /*
     * get_superadmin_id(): Returns the Maia user ID of the
     *                      superadministrator, or 0 if no one has
     *                      superadministrator privileges.
     */
    function get_superadmin_id()
    {
        global $dbh;

        $super_id = 0;
        $select = "SELECT id FROM maia_users WHERE user_level = 'S'";
        $sth = $dbh->query($select);
        if ($row = $sth->fetchRow()) {
            $super_id = $row["id"];
        }
        $sth->free();

        return $super_id;
    }


    /*
     * set_superadmin_id(): Makes the specified user the superadministrator.
     */
    function set_superadmin_id($uid)
    {
        global $dbh;

        if (get_superadmin_id() == 0) {
            $update = "UPDATE maia_users SET user_level = 'S' " .
                      "WHERE id = ?";
            $dbh->query($update, array($uid));
            return true;
        } else {
            return false;
        }
    }


    /*
     * unset_superadmin_id(): Demotes the current superadministrator to
     *                        regular (U)ser status.
     */
    function unset_superadmin_id($uid)
    {
        global $dbh;

        $super_id = get_superadmin_id();
        if ($super_id == $uid) {
            $update = "UPDATE maia_users SET user_level = 'U' " .
                      "WHERE id = ?";
            $dbh->query($update, array($uid));
            return true;
        } else {
            return false;
        }
    }


    /*
     * is_system_default_user(): Returns true if the given user
     *                           is the system's pseudo-user.
     */
    function is_system_default_user($uid)
    {
        return (get_user_name($uid) == "@.");
    }


    /*
     * is_a_domain_default_user(): Returns true if the given user
     *                             is a domain-class pseudo-user.
     */
    function is_a_domain_default_user($uid)
    {
    	$user_name = get_user_name($uid);
    	return ($user_name[0] == "@");
    }


    /*
     * add_policy(): Creates a new policy record for a given
     *               e-mail address.
     */
    function add_policy($email)
    {
        global $dbh;

        $policy_id = 0;
        $domain = "@" . get_domain_from_email($email);
        $select = "SELECT virus_lover, " .
                         "spam_lover, " .
                         "banned_files_lover, " .
                         "bad_header_lover, " .
                         "bypass_virus_checks, " .
                         "bypass_spam_checks, " .
                         "bypass_banned_checks, " .
                         "bypass_header_checks, " .
                         "discard_viruses, " .
                         "discard_spam, " .
                         "discard_banned_files, " .
                         "discard_bad_headers, " ."spam_modifies_subj, " .
                         "spam_tag_level, " .
                         "spam_tag2_level, " .
                         "spam_kill_level " .
                  "FROM policy WHERE policy_name = ?";

        // Try to find a domain-based set of defaults for this user,
        // based on his e-mail address.
        $sth = $dbh->query($select, array($domain));
        if ($row = $sth->fetchRow()) {
            $virus_lover = $row["virus_lover"];
            $spam_lover = $row["spam_lover"];
            $bad_header_lover = $row["bad_header_lover"];
            $banned_files_lover = $row["banned_files_lover"];
            $bypass_virus_checks = $row["bypass_virus_checks"];
            $bypass_spam_checks = $row["bypass_spam_checks"];
            $bypass_banned_checks = $row["bypass_banned_checks"];
            $bypass_header_checks = $row["bypass_header_checks"];
            $discard_viruses = $row["discard_viruses"];
            $discard_spam = $row["discard_spam"];
            $discard_banned_files = $row["discard_banned_files"];
            $discard_bad_headers = $row["discard_bad_headers"];
            $spam_modifies_subj = $row["spam_modifies_subj"];
            $spam_tag_level = $row["spam_tag_level"];
            $spam_tag2_level = $row["spam_tag2_level"];
            $spam_kill_level = $row["spam_kill_level"];
            $nodefault = false;
        } else {
            $sth->free();

            // Try to find a "Default" policy (@.) to copy defaults from.
            $sth = $dbh->query($select, array("Default"));
            if ($row = $sth->fetchRow()) {
                $virus_lover = $row["virus_lover"];
                $spam_lover = $row["spam_lover"];
                $bad_header_lover = $row["bad_header_lover"];
                $banned_files_lover = $row["banned_files_lover"];
                $bypass_virus_checks = $row["bypass_virus_checks"];
                $bypass_spam_checks = $row["bypass_spam_checks"];
                $bypass_banned_checks = $row["bypass_banned_checks"];
                $bypass_header_checks = $row["bypass_header_checks"];
                $discard_viruses = $row["discard_viruses"];
                $discard_spam = $row["discard_spam"];
                $discard_banned_files = $row["discard_banned_files"];
                $discard_bad_headers = $row["discard_bad_headers"];
                $spam_modifies_subj = $row["spam_modifies_subj"];
                $spam_tag_level = $row["spam_tag_level"];
                $spam_tag2_level = $row["spam_tag2_level"];
                $spam_kill_level = $row["spam_kill_level"];
                $nodefault = false;
            } else {

            	// No suitable defaults found.
            	$nodefault = true;
            }
        }
        $sth->free();

        if ($nodefault) {

            // Use the database defaults as our last resort.
            $insert = "INSERT INTO policy (policy_name) VALUES (?)";
            $dbh->query($insert, array($email));

        } else {

            // Use the domain or system default values found above.
            $insert = "INSERT INTO policy (policy_name, " .
                                          "virus_lover, " .
                                          "spam_lover, " .
                                          "banned_files_lover, " .
                                          "bad_header_lover, " .
                                          "bypass_virus_checks, " .
                                          "bypass_spam_checks, " .
                                          "bypass_banned_checks, " .
                                          "bypass_header_checks, " .
                                          "discard_viruses, " .
                                          "discard_spam, " .
                                          "discard_banned_files, " .
                                          "discard_bad_headers, " .
                                          "spam_modifies_subj, " .
                                          "spam_tag_level, " .
                                          "spam_tag2_level, " .
                                          "spam_kill_level" .
                      ") VALUES (?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?)";
            $dbh->query($insert, array($email,
                                       $virus_lover,
                                       $spam_lover,
                                       $banned_files_lover,
                                       $bad_header_lover,
                                       $bypass_virus_checks,
                                       $bypass_spam_checks,
                                       $bypass_banned_checks,
                                       $bypass_header_checks,
                                       $discard_viruses,
                                       $discard_spam,
                                       $discard_banned_files,
                                       $discard_bad_headers,
                                       $spam_modifies_subj,
                                       $spam_tag_level,
                                       $spam_tag2_level,
                                       $spam_kill_level));
        }

        $select = "SELECT id FROM policy WHERE policy_name = ?";
        $sth = $dbh->query($select, array($email));
        if ($row = $sth->fetchRow()) {
            $policy_id = $row["id"];
        }
        $sth->free();

        return $policy_id;
    }
    
    /*
     *
     * get_autocreation_policy():  Gets autocreation policy for given domain 
     *                Returns 'Y' or 'N', or false on error.
     */
    function get_autocreation_policy($domain) {
      global $dbh;
      
      $query = "SELECT enable_user_autocreation FROM maia_domains WHERE domain = ?";
      $sth = $dbh->query($query, array($domain));
      if ($row = $sth->fetchrow()) {
        $policy = $row["enable_user_autocreation"];
      } else {
        $policy = false;
      }
      $sth->free();
      return $policy;
    }
    
     /*
     *
     * get_autocreation_policy():  Sets autocreation policy for given domain 
     *               
     */
    function set_autocreation_policy($domain, $policy) {
      global $dbh;
      
      $query = "UPDATE maia_domains SET enable_user_autocreation=? FROM maia_domains WHERE domain = ?";
      $sth = $dbh->query($query, array($policy, $domain));
    }
    
    /*
     *
     * add_domain():  adds a domain default account. 
     *                Copies defaults in from system default.
     */
    function add_domain($domain_name) {
      global $dbh;
      
      // get default autocreation policy
      $default_autocreation_policy = get_autocreation_policy("@.");

      // Add the domain to the maia_domains table
      $insert = "INSERT INTO maia_domains (domain, enable_user_autocreation) VALUES (?,?)";
      $dbh->query($insert, array($domain_name, $default_autocreation_policy));
      $select = "SELECT id FROM maia_domains WHERE domain = ?";
      $sth = $dbh->query($select, array($domain_name));
      if ($row = $sth->fetchrow()) {
          $domain_id = $row["id"];
      }
      $sth->free();

      // Add a new policy for the domain, based on the
      // system defaults.
      $policy_id = add_policy($domain_name);
      // Add the domain address to the users table
      $primary_email_id = add_address_to_user($policy_id, $domain_name, 0, $domain_id);

      $default_user_config = get_maia_user_row(get_user_id("@.", "@."));
      
      // Add a domain default user to the maia_users table
      $insert = "INSERT INTO maia_users (user_name, primary_email_id, reminders, discard_ham, theme_id) VALUES (?, ?, 'N', ?, ?)";
      $dbh->query($insert, array($domain_name, 
                                 $primary_email_id,
                                 $default_user_config["discard_ham"],
                                 $default_user_config["theme_id"]
                                 ));
      $select = "SELECT id FROM maia_users WHERE user_name = ?";
      $sth = $dbh->query($select, array($domain_name));
      if ($row = $sth->fetchrow()) {
          $maia_user_id = $row["id"];
      }
      $sth->free();

      // Update the users table to link the e-mail address back to the domain
      $update = "UPDATE users SET maia_user_id = ? WHERE id = ?";
      $dbh->query($update, array($maia_user_id, $primary_email_id));

   
      
    
    } 
    
    /*
     * add_address_to_user(): Add a address to the users table
     *
     */
    function add_address_to_user($policy_id, $address, $uid, $domain_id)
    {
      global $dbh;
      $priority = get_email_address_priority($address);
      $insert = "INSERT INTO users (policy_id, email, priority, maia_user_id, maia_domain_id) VALUES (?,?,?,?,?)";
      $dbh->query($insert, array($policy_id, $address, $priority, $uid, $domain_id));
      $select = "SELECT id FROM users WHERE email = ?";
      $sth = $dbh->query($select, array($address));
      if ($row = $sth->fetchrow()) {
          $email_id = $row["id"];
      } else {
        $email_id = 0;
      }
      $sth->free();
      return $email_id;
    }
     

    /*
     * add_email_address_to_user(): Links a new e-mail address to the
     *                              specified user's account.
     */
    function add_email_address_to_user($uid, $email)
    {
    	global $dbh;

  	// Add a new policy for this user
  	$policy_id = add_policy($email);

  	// Add the user's e-mail address to the amavisd users table
    $new_address_id = add_address_to_user($policy_id, $email, $uid, 0);
    return $new_address_id;
    }


    /*
     * get_primary_email_id(): Returns the ID of the specified user's
     *                         primary e-mail address.
     */
    function get_primary_email_id($user_id)
    {
        global $dbh;

        $email_id = 0;
        $select = "SELECT primary_email_id FROM maia_users WHERE id = ?";
        $sth = $dbh->query($select, array($user_id));
        if ($row = $sth->fetchrow()) {
            $email_id = $row["primary_email_id"];
        }
        $sth->free();

        return $email_id;
    }


    /*
     * set_primary_email(): Registers the specified e-mail address
     *                      as the specified user's primary address.
     */
    function set_primary_email($user_id, $email_id)
    {
        global $dbh;

        $update = "UPDATE maia_users SET primary_email_id = ? WHERE id = ?";
        $dbh->query($update, array($email_id, $user_id));
    }


    /*
     * add_user(): Creates a new user record, based on a new e-mail
     *             address and user name.
     */
    function add_user($user_name, $email)
    {
    	global $dbh;

        // get domain default if available.... 
        $domain = get_domain_from_email($email);
        $domain_id = get_user_id("@" . $domain, "@" . $domain);
        if ($domain_id != 0) {
          $domain_defaults = get_maia_user_row($domain_id);
        } else {
          $domain_defaults = get_maia_user_row(get_user_id("@.", "@."));
        }
        
        
        // Add an entry to the maia_users table
        $insert = "INSERT INTO maia_users (user_name, theme_id) VALUES (?,?)";
        $dbh->query($insert, array($user_name, $domain_defaults["theme_id"]));

        // Get the UID of this new record
        $select = "SELECT id FROM maia_users WHERE user_name = ?";
        $sth = $dbh->query($select, array($user_name));
        if ($row = $sth->fetchRow()) {
            $uid = $row["id"];
        }
        $sth->free();

        // Link this e-mail address to this user
        $email_id = add_email_address_to_user($uid, $email);

        // Make this e-mail address the user's primary address
        set_primary_email($uid, $email_id);

        return $uid;
    }


    /*
     * get_email_address_by_id(): Look up an e-mail address by
     *                            its ID.
     */
    function get_email_address_by_id($address_id)
    {
        global $dbh;

        $email = "";
        $select = "SELECT email FROM users WHERE id = ?";
        $sth = $dbh->query($select, array($address_id));
        if ($row = $sth->fetchrow()) {
            $email = $row["email"];
        }
        $sth->free();

        return $email;
    }


    /*
     * get_email_address_id(): Look up the ID of an e-mail address.
     */
    function get_email_address_id($address)
    {
        global $dbh;

        $address_id = 0;
        $select = "SELECT id FROM users WHERE email = ?";
        $sth = $dbh->query($select, array($address));
        if ($row = $sth->fetchRow()) {
            $address_id = $row["id"];
        }
        $sth->free();

        return $address_id;
    }


    /*
     * get_email_address_owner(): Returns the ID of the user that
     *                            owns the specified e-mail address.
     */
    function get_email_address_owner($address_id)
    {
        global $dbh;

        $owner_id = 0;
        $select = "SELECT maia_user_id FROM users WHERE id = ?";
        $sth = $dbh->query($select, array($address_id));
        if ($row = $sth->fetchRow()) {
            $owner_id = $row["maia_user_id"];
        }
        $sth->free();

        return $owner_id;
    }


    /*
     * transfer_email_address_to_user(): Transfers an e-mail address
     *                                   from one user to another.
     */
    function transfer_email_address_to_user($old_owner_id, $new_owner_id, $email)
    {
    	global $dbh;

    	// Link the e-mail address to the new owner
    	$update = "UPDATE users SET maia_user_id = ? WHERE email = ?";
    	$dbh->query($update, array($new_owner_id, $email));

    	// If the old owner has no other e-mail addresses, merge
    	// the old owner's assets into the new owner's account
    	// and delete the old owner.
    	$select = "SELECT id FROM users WHERE maia_user_id = ?";
    	$sth = $dbh->query($select, array($old_owner_id));
    	if (!$sth->fetchRow()) {
    	    transfer_mail_to_user($old_owner_id, $new_owner_id);
    	    transfer_wblist_to_user($old_owner_id, $new_owner_id);
    	    transfer_domain_admin_to_user($old_owner_id, $new_owner_id);
    	    transfer_stats_to_user($old_owner_id, $new_owner_id);
            update_mail_stats($new_owner_id, "suspected_ham");
            update_mail_stats($new_owner_id, "suspected_spam");
    	    delete_user($old_owner_id);
    	}
    	$sth->free();
    }


    /*
     * transfer_stats_to_user(): Merges one user's mail statistics
     *                           into another's.
     */
    function transfer_stats_to_user($old_owner_id, $new_owner_id)
    {
        global $dbh;

        $select = "SELECT oldest_ham_date, " .
                         "newest_ham_date, " .
                         "smallest_ham_size, " .
                         "largest_ham_size, " .
                         "total_ham_size, " .
                         "lowest_ham_score, " .
                         "highest_ham_score, " .
                         "total_ham_score, " .
                         "total_ham_items, " .
                         "oldest_wl_date, " .
                         "newest_wl_date, " .
                         "smallest_wl_size, " .
                         "largest_wl_size, " .
                         "total_wl_size, " .
                         "total_wl_items, " .
                         "oldest_bl_date, " .
                         "newest_bl_date, " .
                         "smallest_bl_size, " .
                         "largest_bl_size, " .
                         "total_bl_size, " .
                         "total_bl_items, " .
                         "oldest_fp_date, " .
                         "newest_fp_date, " .
                         "smallest_fp_size, " .
                         "largest_fp_size, " .
                         "total_fp_size, " .
                         "lowest_fp_score, " .
                         "highest_fp_score, " .
                         "total_fp_score, " .
                         "total_fp_items, " .
                         "oldest_fn_date, " .
                         "newest_fn_date, " .
                         "smallest_fn_size, " .
                         "largest_fn_size, " .
                         "total_fn_size, " .
                         "lowest_fn_score, " .
                         "highest_fn_score, " .
                         "total_fn_score, " .
                         "total_fn_items, " .
                         "oldest_spam_date, " .
                         "newest_spam_date, " .
                         "smallest_spam_size, " .
                         "largest_spam_size, " .
                         "total_spam_size, " .
                         "lowest_spam_score, " .
                         "highest_spam_score, " .
                         "total_spam_score, " .
                         "total_spam_items, " .
                         "oldest_virus_date, " .
                         "newest_virus_date, " .
                         "smallest_virus_size, " .
                         "largest_virus_size, " .
                         "total_virus_size, " .
                         "total_virus_items, " .
                         "oldest_bad_header_date, " .
                         "newest_bad_header_date, " .
                         "smallest_bad_header_size, " .
                         "largest_bad_header_size, " .
                         "total_bad_header_size, " .
                         "total_bad_header_items, " .
                         "oldest_banned_file_date, " .
                         "newest_banned_file_date, " .
                         "smallest_banned_file_size, " .
                         "largest_banned_file_size, " .
                         "total_banned_file_size, " .
                         "total_banned_file_items, " .
                         "oldest_oversized_date, " .
                         "newest_oversized_date, " .
                         "smallest_oversized_size, " .
                         "largest_oversized_size, " .
                         "total_oversized_size, " .
                         "total_oversized_items " .
                  "FROM maia_stats WHERE user_id = ?";
        $sth = $dbh->query($select, array($old_owner_id));
        if ($row = $sth->fetchRow()) {
            $old_oldest_ham_date = $row["oldest_ham_date"];
            $old_newest_ham_date = $row["newest_ham_date"];
            $old_smallest_ham_size = $row["smallest_ham_size"];
            $old_largest_ham_size = $row["largest_ham_size"];
            $old_total_ham_size = $row["total_ham_size"];
            $old_lowest_ham_score = $row["lowest_ham_score"];
            $old_highest_ham_score = $row["highest_ham_score"];
            $old_total_ham_score = $row["total_ham_score"];
            $old_total_ham_items = $row["total_ham_items"];
            $old_oldest_wl_date = $row["oldest_wl_date"];
            $old_newest_wl_date = $row["newest_wl_date"];
            $old_smallest_wl_size = $row["smallest_wl_size"];
            $old_largest_wl_size = $row["largest_wl_size"];
            $old_total_wl_size = $row["total_wl_size"];
            $old_total_wl_items = $row["total_wl_items"];
            $old_oldest_bl_date = $row["oldest_bl_date"];
            $old_newest_bl_date = $row["newest_bl_date"];
            $old_smallest_bl_size = $row["smallest_bl_size"];
            $old_largest_bl_size = $row["largest_bl_size"];
            $old_total_bl_size = $row["total_bl_size"];
            $old_total_bl_items = $row["total_bl_items"];
            $old_oldest_fp_date = $row["oldest_fp_date"];
            $old_newest_fp_date = $row["newest_fp_date"];
            $old_smallest_fp_size = $row["smallest_fp_size"];
            $old_largest_fp_size = $row["largest_fp_size"];
            $old_total_fp_size = $row["total_fp_size"];
            $old_lowest_fp_score = $row["lowest_fp_score"];
            $old_highest_fp_score = $row["highest_fp_score"];
            $old_total_fp_score = $row["total_fp_score"];
            $old_total_fp_items = $row["total_fp_items"];
            $old_oldest_fn_date = $row["oldest_fn_date"];
            $old_newest_fn_date = $row["newest_fn_date"];
            $old_smallest_fn_size = $row["smallest_fn_size"];
            $old_largest_fn_size = $row["largest_fn_size"];
            $old_total_fn_size = $row["total_fn_size"];
            $old_lowest_fn_score = $row["lowest_fn_score"];
            $old_highest_fn_score = $row["highest_fn_score"];
            $old_total_fn_score = $row["total_fn_score"];
            $old_total_fn_items = $row["total_fn_items"];
            $old_oldest_spam_date = $row["oldest_spam_date"];
            $old_newest_spam_date = $row["newest_spam_date"];
            $old_smallest_spam_size = $row["smallest_spam_size"];
            $old_largest_spam_size = $row["largest_spam_size"];
            $old_total_spam_size = $row["total_spam_size"];
            $old_lowest_spam_score = $row["lowest_spam_score"];
            $old_highest_spam_score = $row["highest_spam_score"];
            $old_total_spam_score = $row["total_spam_score"];
            $old_total_spam_items = $row["total_spam_items"];
            $old_oldest_virus_date = $row["oldest_virus_date"];
            $old_newest_virus_date = $row["newest_virus_date"];
            $old_smallest_virus_size = $row["smallest_virus_size"];
            $old_largest_virus_size = $row["largest_virus_size"];
            $old_total_virus_size = $row["total_virus_size"];
            $old_total_virus_items = $row["total_virus_items"];
            $old_oldest_bad_header_date = $row["oldest_bad_header_date"];
            $old_newest_bad_header_date = $row["newest_bad_header_date"];
            $old_smallest_bad_header_size = $row["smallest_bad_header_size"];
            $old_largest_bad_header_size = $row["largest_bad_header_size"];
            $old_total_bad_header_size = $row["total_bad_header_size"];
            $old_total_bad_header_items = $row["total_bad_header_items"];
            $old_oldest_banned_file_date = $row["oldest_banned_file_date"];
            $old_newest_banned_file_date = $row["newest_banned_file_date"];
            $old_smallest_banned_file_size = $row["smallest_banned_file_size"];
            $old_largest_banned_file_size = $row["largest_banned_file_size"];
            $old_total_banned_file_size = $row["total_banned_file_size"];
            $old_total_banned_file_items = $row["total_banned_file_items"];
            $old_oldest_oversized_date = $row["oldest_oversized_date"];
            $old_newest_oversized_date = $row["newest_oversized_date"];
            $old_smallest_oversized_size = $row["smallest_oversized_size"];
            $old_largest_oversized_size = $row["largest_oversized_size"];
            $old_total_oversized_size = $row["total_oversized_size"];
            $old_total_oversized_items = $row["total_oversized_items"];

            // New owner already has a stats record, merge the old owner's
            // stats into this record appropriately.
            $sth2 = $dbh->query($select, array($new_owner_id));
            if ($row2 = $sth2->fetchRow()) {

            	// Confirmed Ham
                $new_oldest_ham_date = $row2["oldest_ham_date"];
                if ($old_oldest_ham_date != NULL && $old_oldest_ham_date < $new_oldest_ham_date) {
                    $new_oldest_ham_date = $old_oldest_ham_date;
                }
                $new_newest_ham_date = $row2["newest_ham_date"];
                if ($old_newest_ham_date > $new_newest_ham_date) {
                    $new_newest_ham_date = $old_newest_ham_date;
                }
                $new_smallest_ham_size = $row2["smallest_ham_size"];
                if ($old_smallest_ham_size < $new_smallest_ham_size) {
                    $new_smallest_ham_size = $old_smallest_ham_size;
                }
                $new_largest_ham_size = $row2["largest_ham_size"];
                if ($old_largest_ham_size > $new_largest_ham_size) {
                    $new_largest_ham_size = $old_largest_ham_size;
                }
                $new_total_ham_size = $row2["total_ham_size"] + $old_total_ham_size;
                $new_lowest_ham_score = $row2["lowest_ham_score"];
                if ($old_lowest_ham_score < $new_lowest_ham_score) {
                    $new_lowest_ham_score = $old_lowest_ham_score;
                }
                $new_highest_ham_score = $row2["highest_ham_score"];
                if ($old_highest_ham_score > $new_highest_ham_score) {
                    $new_highest_ham_score = $old_highest_ham_score;
                }
                $new_total_ham_score = $row2["total_ham_score"] + $old_total_ham_score;
                $new_total_ham_items = $row2["total_ham_items"] + $old_total_ham_items;

                // Whitelisted Sender
                $new_oldest_wl_date = $row2["oldest_wl_date"];
                if ($old_oldest_wl_date != NULL && $old_oldest_wl_date < $new_oldest_wl_date) {
                    $new_oldest_wl_date = $old_oldest_wl_date;
                }
                $new_newest_wl_date = $row2["newest_wl_date"];
                if ($old_newest_wl_date > $new_newest_wl_date) {
                    $new_newest_wl_date = $old_newest_wl_date;
                }
                $new_smallest_wl_size = $row2["smallest_wl_size"];
                if ($old_smallest_wl_size < $new_smallest_wl_size) {
                    $new_smallest_wl_size = $old_smallest_wl_size;
                }
                $new_largest_wl_size = $row2["largest_wl_size"];
                if ($old_largest_wl_size > $new_largest_wl_size) {
                    $new_largest_wl_size = $old_largest_wl_size;
                }
                $new_total_wl_size = $row2["total_wl_size"] + $old_total_wl_size;
                $new_total_wl_items = $row2["total_wl_items"] + $old_total_wl_items;

                // Blacklisted Sender
                $new_oldest_bl_date = $row2["oldest_bl_date"];
                if ($old_oldest_bl_date != NULL && $old_oldest_bl_date < $new_oldest_bl_date) {
                    $new_oldest_bl_date = $old_oldest_bl_date;
                }
                $new_newest_bl_date = $row2["newest_bl_date"];
                if ($old_newest_bl_date > $new_newest_bl_date) {
                    $new_newest_bl_date = $old_newest_bl_date;
                }
                $new_smallest_bl_size = $row2["smallest_bl_size"];
                if ($old_smallest_bl_size < $new_smallest_bl_size) {
                    $new_smallest_bl_size = $old_smallest_bl_size;
                }
                $new_largest_bl_size = $row2["largest_bl_size"];
                if ($old_largest_bl_size > $new_largest_bl_size) {
                    $new_largest_bl_size = $old_largest_bl_size;
                }
                $new_total_bl_size = $row2["total_bl_size"] + $old_total_bl_size;
                $new_total_bl_items = $row2["total_bl_items"] + $old_total_bl_items;

                // False Positives
                $new_oldest_fp_date = $row2["oldest_fp_date"];
                if ($old_oldest_fp_date != NULL && $old_oldest_fp_date < $new_oldest_fp_date) {
                    $new_oldest_fp_date = $old_oldest_fp_date;
                }
                $new_newest_fp_date = $row2["newest_fp_date"];
                if ($old_newest_fp_date > $new_newest_fp_date) {
                    $new_newest_fp_date = $old_newest_fp_date;
                }
                $new_smallest_fp_size = $row2["smallest_fp_size"];
                if ($old_smallest_fp_size < $new_smallest_fp_size) {
                    $new_smallest_fp_size = $old_smallest_fp_size;
                }
                $new_largest_fp_size = $row2["largest_fp_size"];
                if ($old_largest_fp_size > $new_largest_fp_size) {
                    $new_largest_fp_size = $old_largest_fp_size;
                }
                $new_total_fp_size = $row2["total_fp_size"] + $old_total_fp_size;
                $new_lowest_fp_score = $row2["lowest_fp_score"];
                if ($old_lowest_fp_score < $new_lowest_fp_score) {
                    $new_lowest_fp_score = $old_lowest_fp_score;
                }
                $new_highest_fp_score = $row2["highest_fp_score"];
                if ($old_highest_fp_score > $new_highest_fp_score) {
                    $new_highest_fp_score = $old_highest_fp_score;
                }
                $new_total_fp_score = $row2["total_fp_score"] + $old_total_fp_score;
                $new_total_fp_items = $row2["total_fp_items"] + $old_total_fp_items;

                // False Negatives
                $new_oldest_fn_date = $row2["oldest_fn_date"];
                if ($old_oldest_fn_date != NULL && $old_oldest_fn_date < $new_oldest_fn_date) {
                    $new_oldest_fn_date = $old_oldest_fn_date;
                }
                $new_newest_fn_date = $row2["newest_fn_date"];
                if ($old_newest_fn_date > $new_newest_fn_date) {
                    $new_newest_fn_date = $old_newest_fn_date;
                }
                $new_smallest_fn_size = $row2["smallest_fn_size"];
                if ($old_smallest_fn_size < $new_smallest_fn_size) {
                    $new_smallest_fn_size = $old_smallest_fn_size;
                }
                $new_largest_fn_size = $row2["largest_fn_size"];
                if ($old_largest_fn_size > $new_largest_fn_size) {
                    $new_largest_fn_size = $old_largest_fn_size;
                }
                $new_total_fn_size = $row2["total_fn_size"] + $old_total_fn_size;
                $new_lowest_fn_score = $row2["lowest_fn_score"];
                if ($old_lowest_fn_score < $new_lowest_fn_score) {
                    $new_lowest_fn_score = $old_lowest_fn_score;
                }
                $new_highest_fn_score = $row2["highest_fn_score"];
                if ($old_highest_fn_score > $new_highest_fn_score) {
                    $new_highest_fn_score = $old_highest_fn_score;
                }
                $new_total_fn_score = $row2["total_fn_score"] + $old_total_fn_score;
                $new_total_fn_items = $row2["total_fn_items"] + $old_total_fn_items;

                // Confirmed Spam
                $new_oldest_spam_date = $row2["oldest_spam_date"];
                if ($old_oldest_spam_date != NULL && $old_oldest_spam_date < $new_oldest_spam_date) {
                    $new_oldest_spam_date = $old_oldest_spam_date;
                }
                $new_newest_spam_date = $row2["newest_spam_date"];
                if ($old_newest_spam_date > $new_newest_spam_date) {
                    $new_newest_spam_date = $old_newest_spam_date;
                }
                $new_smallest_spam_size = $row2["smallest_spam_size"];
                if ($old_smallest_spam_size < $new_smallest_spam_size) {
                    $new_smallest_spam_size = $old_smallest_spam_size;
                }
                $new_largest_spam_size = $row2["largest_spam_size"];
                if ($old_largest_spam_size > $new_largest_spam_size) {
                    $new_largest_spam_size = $old_largest_spam_size;
                }
                $new_total_spam_size = $row2["total_spam_size"] + $old_total_spam_size;
                $new_lowest_spam_score = $row2["lowest_spam_score"];
                if ($old_lowest_spam_score < $new_lowest_spam_score) {
                    $new_lowest_spam_score = $old_lowest_spam_score;
                }
                $new_highest_spam_score = $row2["highest_spam_score"];
                if ($old_highest_spam_score > $new_highest_spam_score) {
                    $new_highest_spam_score = $old_highest_spam_score;
                }
                $new_total_spam_score = $row2["total_spam_score"] + $old_total_spam_score;
                $new_total_spam_items = $row2["total_spam_items"] + $old_total_spam_items;

                // Viruses
                $new_oldest_virus_date = $row2["oldest_virus_date"];
                if ($old_oldest_virus_date != NULL && $old_oldest_virus_date < $new_oldest_virus_date) {
                    $new_oldest_virus_date = $old_oldest_virus_date;
                }
                $new_newest_virus_date = $row2["newest_virus_date"];
                if ($old_newest_virus_date > $new_newest_virus_date) {
                    $new_newest_virus_date = $old_newest_virus_date;
                }
                $new_smallest_virus_size = $row2["smallest_virus_size"];
                if ($old_smallest_virus_size < $new_smallest_virus_size) {
                    $new_smallest_virus_size = $old_smallest_virus_size;
                }
                $new_largest_virus_size = $row2["largest_virus_size"];
                if ($old_largest_virus_size > $new_largest_virus_size) {
                    $new_largest_virus_size = $old_largest_virus_size;
                }
                $new_total_virus_size = $row2["total_virus_size"] + $old_total_virus_size;
                $new_total_virus_items = $row2["total_virus_items"] + $old_total_virus_items;

                // Bad Headers
                $new_oldest_bad_header_date = $row2["oldest_bad_header_date"];
                if ($old_oldest_bad_header_date != NULL && $old_oldest_bad_header_date < $new_oldest_bad_header_date) {
                    $new_oldest_bad_header_date = $old_oldest_bad_header_date;
                }
                $new_newest_bad_header_date = $row2["newest_bad_header_date"];
                if ($old_newest_bad_header_date > $new_newest_bad_header_date) {
                    $new_newest_bad_header_date = $old_newest_bad_header_date;
                }
                $new_smallest_bad_header_size = $row2["smallest_bad_header_size"];
                if ($old_smallest_bad_header_size < $new_smallest_bad_header_size) {
                    $new_smallest_bad_header_size = $old_smallest_bad_header_size;
                }
                $new_largest_bad_header_size = $row2["largest_bad_header_size"];
                if ($old_largest_bad_header_size > $new_largest_bad_header_size) {
                    $new_largest_bad_header_size = $old_largest_bad_header_size;
                }
                $new_total_bad_header_size = $row2["total_bad_header_size"] + $old_total_bad_header_size;
                $new_total_bad_header_items = $row2["total_bad_header_items"] + $old_total_bad_header_items;

                // Banned Files
                $new_oldest_banned_file_date = $row2["oldest_banned_file_date"];
                if ($old_oldest_banned_file_date != NULL && $old_oldest_banned_file_date < $new_oldest_banned_file_date) {
                    $new_oldest_banned_file_date = $old_oldest_banned_file_date;
                }
                $new_newest_banned_file_date = $row2["newest_banned_file_date"];
                if ($old_newest_banned_file_date > $new_newest_banned_file_date) {
                    $new_newest_banned_file_date = $old_newest_banned_file_date;
                }
                $new_smallest_banned_file_size = $row2["smallest_banned_file_size"];
                if ($old_smallest_banned_file_size < $new_smallest_banned_file_size) {
                    $new_smallest_banned_file_size = $old_smallest_banned_file_size;
                }
                $new_largest_banned_file_size = $row2["largest_banned_file_size"];
                if ($old_largest_banned_file_size > $new_largest_banned_file_size) {
                    $new_largest_banned_file_size = $old_largest_banned_file_size;
                }
                $new_total_banned_file_size = $row2["total_banned_file_size"] + $old_total_banned_file_size;
                $new_total_banned_file_items = $row2["total_banned_file_items"] + $old_total_banned_file_items;

                // Oversized Items
                $new_oldest_oversized_date = $row2["oldest_oversized_date"];
                if ($old_oldest_oversized_date != NULL && $old_oldest_oversized_date < $new_oldest_oversized_date) {
                    $new_oldest_oversized_date = $old_oldest_oversized_date;
                }
                $new_newest_oversized_date = $row2["newest_oversized_date"];
                if ($old_newest_oversized_date > $new_newest_oversized_date) {
                    $new_newest_oversized_date = $old_newest_oversized_date;
                }
                $new_smallest_oversized_size = $row2["smallest_oversized_size"];
                if ($old_smallest_oversized_size < $new_smallest_oversized_size) {
                    $new_smallest_oversized_size = $old_smallest_oversized_size;
                }
                $new_largest_oversized_size = $row2["largest_oversized_size"];
                if ($old_largest_oversized_size > $new_largest_oversized_size) {
                    $new_largest_oversized_size = $old_largest_oversized_size;
                }
                $new_total_oversized_size = $row2["total_oversized_size"] + $old_total_oversized_size;
                $new_total_oversized_items = $row2["total_oversized_items"] + $old_total_oversized_items;

                $update = "UPDATE maia_stats SET oldest_ham_date = ?, " .
                                                "newest_ham_date = ?, " .
                                                "smallest_ham_size = ?, " .
                                                "largest_ham_size = ?, " .
                                                "total_ham_size = ?, " .
                                                "lowest_ham_score = ?, " .
                                                "highest_ham_score = ?, " .
                                                "total_ham_score = ?, " .
                                                "total_ham_items = ?, " .
                                                "oldest_wl_date = ?, " .
                                                "newest_wl_date = ?, " .
                                                "smallest_wl_size = ?, " .
                                                "largest_wl_size = ?, " .
                                                "total_wl_size = ?, " .
                                                "total_wl_items = ?, " .
                                                "oldest_bl_date = ?, " .
                                                "newest_bl_date = ?, " .
                                                "smallest_bl_size = ?, " .
                                                "largest_bl_size = ?, " .
                                                "total_bl_size = ?, " .
                                                "total_bl_items = ?, " .
                                                "oldest_fp_date = ?, " .
                                                "newest_fp_date = ?, " .
                                                "smallest_fp_size = ?, " .
                                                "largest_fp_size = ?, " .
                                                "total_fp_size = ?, " .
                                                "lowest_fp_score = ?, " .
                                                "highest_fp_score = ?, " .
                                                "total_fp_score = ?, " .
                                                "total_fp_items = ?, " .
                                                "oldest_fn_date = ?, " .
                                                "newest_fn_date = ?, " .
                                                "smallest_fn_size = ?, " .
                                                "largest_fn_size = ?, " .
                                                "total_fn_size = ?, " .
                                                "lowest_fn_score = ?, " .
                                                "highest_fn_score = ?, " .
                                                "total_fn_score = ?, " .
                                                "total_fn_items = ?, " .
                                                "oldest_spam_date = ?, " .
                                                "newest_spam_date = ?, " .
                                                "smallest_spam_size = ?, " .
                                                "largest_spam_size = ?, " .
                                                "total_spam_size = ?, " .
                                                "lowest_spam_score = ?, " .
                                                "highest_spam_score = ?, " .
                                                "total_spam_score = ?, " .
                                                "total_spam_items = ?, " .
                                                "oldest_virus_date = ?, " .
                                                "newest_virus_date = ?, " .
                                                "smallest_virus_size = ?, " .
                                                "largest_virus_size = ?, " .
                                                "total_virus_size = ?, " .
                                                "total_virus_items = ?, " .
                                                "oldest_bad_header_date = ?, " .
                                                "newest_bad_header_date = ?, " .
                                                "smallest_bad_header_size = ?, " .
                                                "largest_bad_header_size = ?, " .
                                                "total_bad_header_size = ?, " .
                                                "total_bad_header_items = ?, " .
                                                "oldest_banned_file_date = ?, " .
                                                "newest_banned_file_date = ?, " .
                                                "smallest_banned_file_size = ?, " .
                                                "largest_banned_file_size = ?, " .
                                                "total_banned_file_size = ?, " .
                                                "total_banned_file_items = ?, " .
                                                "oldest_oversized_date = ?, " .
                                                "newest_oversized_date = ?, " .
                                                "smallest_oversized_size = ?, " .
                                                "largest_oversized_size = ?, " .
                                                "total_oversized_size = ?, " .
                                                "total_oversized_items = ? " .
                          "WHERE user_id = ?";
                $dbh->query($update, array($new_oldest_ham_date,
                                           $new_newest_ham_date,
                                           $new_smallest_ham_size,
                                           $new_largest_ham_size,
                                           $new_total_ham_size,
                                           $new_lowest_ham_score,
                                           $new_highest_ham_score,
                                           $new_total_ham_score,
                                           $new_total_ham_items,
                                           $new_oldest_wl_date,
                                           $new_newest_wl_date,
                                           $new_smallest_wl_size,
                                           $new_largest_wl_size,
                                           $new_total_wl_size,
                                           $new_total_wl_items,
                                           $new_oldest_bl_date,
                                           $new_newest_bl_date,
                                           $new_smallest_bl_size,
                                           $new_largest_bl_size,
                                           $new_total_bl_size,
                                           $new_total_bl_items,
                                           $new_oldest_fp_date,
                                           $new_newest_fp_date,
                                           $new_smallest_fp_size,
                                           $new_largest_fp_size,
                                           $new_total_fp_size,
                                           $new_lowest_fp_score,
                                           $new_highest_fp_score,
                                           $new_total_fp_score,
                                           $new_total_fp_items,
                                           $new_oldest_fn_date,
                                           $new_newest_fn_date,
                                           $new_smallest_fn_size,
                                           $new_largest_fn_size,
                                           $new_total_fn_size,
                                           $new_lowest_fn_score,
                                           $new_highest_fn_score,
                                           $new_total_fn_score,
                                           $new_total_fn_items,
                                           $new_oldest_spam_date,
                                           $new_newest_spam_date,
                                           $new_smallest_spam_size,
                                           $new_largest_spam_size,
                                           $new_total_spam_size,
                                           $new_lowest_spam_score,
                                           $new_highest_spam_score,
                                           $new_total_spam_score,
                                           $new_total_spam_items,
                                           $new_oldest_virus_date,
                                           $new_newest_virus_date,
                                           $new_smallest_virus_size,
                                           $new_largest_virus_size,
                                           $new_total_virus_size,
                                           $new_total_virus_items,
                                           $new_oldest_bad_header_date,
                                           $new_newest_bad_header_date,
                                           $new_smallest_bad_header_size,
                                           $new_largest_bad_header_size,
                                           $new_total_bad_header_size,
                                           $new_total_bad_header_items,
                                           $new_oldest_banned_file_date,
                                           $new_newest_banned_file_date,
                                           $new_smallest_banned_file_size,
                                           $new_largest_banned_file_size,
                                           $new_total_banned_file_size,
                                           $new_total_banned_file_items,
                                           $new_oldest_oversized_date,
                                           $new_newest_oversized_date,
                                           $new_smallest_oversized_size,
                                           $new_largest_oversized_size,
                                           $new_total_oversized_size,
                                           $new_total_oversized_items,
                                           $new_owner_id));

            // New owner does NOT have a stats record yet, so just
            // copy the old owner's stats over directly.
            } else {

                $insert = "INSERT INTO maia_stats (oldest_ham_date, " .
                                                  "newest_ham_date, " .
                                                  "smallest_ham_size, " .
                                                  "largest_ham_size, " .
                                                  "total_ham_size, " .
                                                  "lowest_ham_score, " .
                                                  "highest_ham_score, " .
                                                  "total_ham_score, " .
                                                  "total_ham_items, " .
                                                  "oldest_wl_date, " .
                                                  "newest_wl_date, " .
                                                  "smallest_wl_size, " .
                                                  "largest_wl_size, " .
                                                  "total_wl_size, " .
                                                  "total_wl_items, " .
                                                  "oldest_bl_date, " .
                                                  "newest_bl_date, " .
                                                  "smallest_bl_size, " .
                                                  "largest_bl_size, " .
                                                  "total_bl_size, " .
                                                  "total_bl_items, " .
                                                  "oldest_fp_date, " .
                                                  "newest_fp_date, " .
                                                  "smallest_fp_size, " .
                                                  "largest_fp_size, " .
                                                  "total_fp_size, " .
                                                  "lowest_fp_score, " .
                                                  "highest_fp_score, " .
                                                  "total_fp_score, " .
                                                  "total_fp_items, " .
                                                  "oldest_fn_date, " .
                                                  "newest_fn_date, " .
                                                  "smallest_fn_size, " .
                                                  "largest_fn_size, " .
                                                  "total_fn_size, " .
                                                  "lowest_fn_score, " .
                                                  "highest_fn_score, " .
                                                  "total_fn_score, " .
                                                  "total_fn_items, " .
                                                  "oldest_spam_date, " .
                                                  "newest_spam_date, " .
                                                  "smallest_spam_size, " .
                                                  "largest_spam_size, " .
                                                  "total_spam_size, " .
                                                  "lowest_spam_score, " .
                                                  "highest_spam_score, " .
                                                  "total_spam_score, " .
                                                  "total_spam_items, " .
                                                  "oldest_virus_date, " .
                                                  "newest_virus_date, " .
                                                  "smallest_virus_size, " .
                                                  "largest_virus_size, " .
                                                  "total_virus_size, " .
                                                  "total_virus_items, " .
                                                  "oldest_bad_header_date, " .
                                                  "newest_bad_header_date, " .
                                                  "smallest_bad_header_size, " .
                                                  "largest_bad_header_size, " .
                                                  "total_bad_header_size, " .
                                                  "total_bad_header_items, " .
                                                  "oldest_banned_file_date, " .
                                                  "newest_banned_file_date, " .
                                                  "smallest_banned_file_size, " .
                                                  "largest_banned_file_size, " .
                                                  "total_banned_file_size, " .
                                                  "total_banned_file_items, " .
                                                  "oldest_oversized_date, " .
                                                  "newest_oversized_date, " .
                                                  "smallest_oversized_size, " .
                                                  "largest_oversized_size, " .
                                                  "total_oversized_size, " .
                                                  "total_oversized_items, " .
                                                  "user_id) " .
                          "VALUES (?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?," .
                                  "?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?," .
                                  "?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?," .
                                  "?,?,?,?,?,?,?,?,?,?,?,?,?)";
                $dbh->query($insert, array($old_oldest_ham_date,
                                           $old_newest_ham_date,
                                           $old_smallest_ham_size,
                                           $old_largest_ham_size,
                                           $old_total_ham_size,
                                           $old_lowest_ham_score,
                                           $old_highest_ham_score,
                                           $old_total_ham_score,
                                           $old_total_ham_items,
                                           $old_oldest_wl_date,
                                           $old_newest_wl_date,
                                           $old_smallest_wl_size,
                                           $old_largest_wl_size,
                                           $old_total_wl_size,
                                           $old_total_wl_items,
                                           $old_oldest_bl_date,
                                           $old_newest_bl_date,
                                           $old_smallest_bl_size,
                                           $old_largest_bl_size,
                                           $old_total_bl_size,
                                           $old_total_bl_items,
                                           $old_oldest_fp_date,
                                           $old_newest_fp_date,
                                           $old_smallest_fp_size,
                                           $old_largest_fp_size,
                                           $old_total_fp_size,
                                           $old_lowest_fp_score,
                                           $old_highest_fp_score,
                                           $old_total_fp_score,
                                           $old_total_fp_items,
                                           $old_oldest_fn_date,
                                           $old_newest_fn_date,
                                           $old_smallest_fn_size,
                                           $old_largest_fn_size,
                                           $old_total_fn_size,
                                           $old_lowest_fn_score,
                                           $old_highest_fn_score,
                                           $old_total_fn_score,
                                           $old_total_fn_items,
                                           $old_oldest_spam_date,
                                           $old_newest_spam_date,
                                           $old_smallest_spam_size,
                                           $old_largest_spam_size,
                                           $old_total_spam_size,
                                           $old_lowest_spam_score,
                                           $old_highest_spam_score,
                                           $old_total_spam_score,
                                           $old_total_spam_items,
                                           $old_oldest_virus_date,
                                           $old_newest_virus_date,
                                           $old_smallest_virus_size,
                                           $old_largest_virus_size,
                                           $old_total_virus_size,
                                           $old_total_virus_items,
                                           $old_oldest_bad_header_date,
                                           $old_newest_bad_header_date,
                                           $old_smallest_bad_header_size,
                                           $old_largest_bad_header_size,
                                           $old_total_bad_header_size,
                                           $old_total_bad_header_items,
                                           $old_oldest_banned_file_date,
                                           $old_newest_banned_file_date,
                                           $old_smallest_banned_file_size,
                                           $old_largest_banned_file_size,
                                           $old_total_banned_file_size,
                                           $old_total_banned_file_items,
                                           $old_oldest_oversized_date,
                                           $old_newest_oversized_date,
                                           $old_smallest_oversized_size,
                                           $old_largest_oversized_size,
                                           $old_total_oversized_size,
                                           $old_total_oversized_items,
                                           $new_owner_id));
            }
            $sth2->free();
        }
        $sth->free();
    }


    /*
     * transfer_domain_admin_to_user(): Transfer domain administration
     *                                  privileges from one user to another.
     */
    function transfer_domain_admin_to_user($old_owner_id, $new_owner_id)
    {
        global $dbh;

        $update = "UPDATE maia_domain_admins SET admin_id = ? WHERE admin_id = ?";
        $dbh->query($update, array($new_owner_id, $old_owner_id));
    }


    /*
     * transfer_wblist_to_user(): Merge one user's whitelist and blacklist
     *                            into another's.
     */
    function transfer_wblist_to_user($old_owner_id, $new_owner_id)
    {
        global $dbh;

        $update = "UPDATE wblist SET rid = ? WHERE rid = ?";
        $dbh->query($update, array($new_owner_id, $old_owner_id));
    }


    /*
     * transfer_mail_to_user(): Transfer one user's e-mail assets (quarantine
     *                          and ham cache) to another.
     */
    function transfer_mail_to_user($old_owner_id, $new_owner_id)
    {
        global $dbh;

        $update = "UPDATE maia_mail_recipients SET recipient_id = ? WHERE recipient_id = ?";
        $dbh->query($update, array($new_owner_id, $old_owner_id));
    }


    /*
     * delete_user(): Deletes a user and all of his assets.
     */
    function delete_user($uid)
    {
        global $dbh;

        // Delete any e-mail addresses this user owns
        delete_user_email_addresses($uid);

        // Delete the user's statistics
        $delete = "DELETE FROM maia_stats WHERE user_id = ?";
        $dbh->query($delete, array($uid));

        // Delete the user's white/blacklist entries
        delete_user_wb_entries($uid);

        // Delete the user's domain admin records
        $delete = "DELETE FROM maia_domain_admins WHERE admin_id = ?";
        $dbh->query($delete, array($uid));

        // Delete the user's mail references
        delete_user_mail_references($uid);

        // Delete the user's record itself
        $delete = "DELETE FROM maia_users WHERE id = ?";
        $dbh->query($delete, array($uid));
    }


    /*
     * delete_policy(): Deletes the specified policy record.
     */
    function delete_policy($policy_id)
    {
        global $dbh;

        $delete = "DELETE FROM policy WHERE id = ?";
        $dbh->query($delete, array($policy_id));
    }


    /*
     * delete_email_address(): Deletes the specified e-mail address.
     */
    function delete_email_address($address_id)
    {
        global $dbh;

        $select = "SELECT policy_id " .
                  "FROM users " .
                  "WHERE id = ?";
        $sth = $dbh->query($select, array($address_id));
        if ($row = $sth->fetchRow()) {
            delete_policy($row["policy_id"]);
            $delete = "DELETE FROM users WHERE id = ?";
            $dbh->query($delete, array($address_id));
        }
        $sth->free();

    }


    /*
     * delete_user_email_addresses(): Deletes all of the e-mail addresses
     *                                that belong to the specified user.
     */
    function delete_user_email_addresses($uid)
    {
        global $dbh;

        $select = "SELECT id FROM users WHERE maia_user_id = ?";
        $sth = $dbh->query($select, array($uid));
        while ($row = $sth->fetchRow()) {
            delete_email_address($row["id"]);
        }
        $sth->free();
    }


    /*
     * delete_mail(): Deletes a single mail item and all references
     *                to it.
     */
    function delete_mail($mail_id)
    {
        global $dbh;

        // Delete any references to recipients
        $delete = "DELETE FROM maia_mail_recipients WHERE mail_id = ?";
        $dbh->query($delete, array($mail_id));

        // Delete any references to SpamAssassin rules
        $delete = "DELETE FROM maia_sa_rules_triggered WHERE mail_id = ?";
        $dbh->query($delete, array($mail_id));

        // Delete any references to viruses
        $delete = "DELETE FROM maia_viruses_detected WHERE mail_id = ?";
        $dbh->query($delete, array($mail_id));

        // Delete any references to banned file attachments
        $delete = "DELETE FROM maia_banned_attachments_found WHERE mail_id = ?";
        $dbh->query($delete, array($mail_id));

        // Delete the mail item itself
        $delete = "DELETE FROM maia_mail WHERE id = ?";
        $dbh->query($delete, array($mail_id));
    }


    /*
     * delete_mail_references(): Deletes a single mail item's recipient
     *                           references.
     */
    function delete_mail_reference($user_id, $mail_id)
    {
        global $dbh;

        // Delete a specific user's recipient reference to this mail.
        $delete = "DELETE FROM maia_mail_recipients WHERE recipient_id = ? AND mail_id = ?";
        $dbh->query($delete, array($user_id, $mail_id));

        // If there are no other recipients referenced by this mail item,
        // delete it.
        $select = "SELECT recipient_id FROM maia_mail_recipients WHERE mail_id = ?";
        $sth = $dbh->query($select, array($mail_id));
        if (!$sth->fetchRow()) {
            delete_mail($mail_id);
        }
        $sth->free();
    }


    /*
     * delete_user_mail_references(): Deletes all recipient mail references
     *                                for the specified user.
     */
    function delete_user_mail_references($uid)
    {
        global $dbh;

        // Delete recipient references to any mail items this user
        // might have.
        $select = "SELECT mail_id FROM maia_mail_recipients WHERE recipient_id = ?";
        $sth = $dbh->query($select, array($uid));
        while ($row = $sth->fetchRow()) {
            delete_mail_reference($uid, $row["mail_id"]);
        }
        $sth->free();
    }


    /*
     * delete_domain_admin_references(): Deletes all administrator references
     *                                   to the specified domain.
     */
    function delete_domain_admin_references($domain_id)
    {
        global $dbh;

        // List all the administrators that reference the target domain.
        $select = "SELECT admin_id FROM maia_domain_admins WHERE domain_id = ?";
        $sth = $dbh->query($select, array($domain_id));
        while ($row = $sth->fetchrow()) {
            $admin_id = $row["admin_id"];

            // Delete this admin's reference to the target domain.
            $delete = "DELETE FROM maia_domain_admins WHERE domain_id = ? AND admin_id = ?";
            $dbh->query($delete, array($domain_id, $admin_id));

            // Does this admin administer any other domains?
            $select = "SELECT domain_id FROM maia_domain_admins WHERE admin_id = ?";
            $sth2 = $dbh->query($select, array($admin_id));
            if (!$sth2->fetchrow()) {

            	// This admin no longer administers any domains, so
            	// return his privilege level to that of a normal user.
                $update = "UPDATE maia_users SET user_level = 'U' WHERE id = ?";
                $dbh->query($update, array($admin_id));
            }
            $sth2->free();
        }
        $sth->free();
    }


    /*
     * delete_domain(): Deletes the specified domain and all administrator
     *                  references to it.
     */
    function delete_domain($domain_id)
    {
    	global $dbh;

    	// Delete all admin references to this domain.
        delete_domain_admin_references($domain_id);

        // Delete the domain record itself.
        $delete = "DELETE FROM maia_domains WHERE id = ?";
        $dbh->query($delete, array($domain_id));

        // Find and delete the default user records associated with this domain
        $select = "SELECT maia_user_id FROM users WHERE maia_domain_id = ?";
        $sth = $dbh->query($select, array($domain_id));
        if ($row = $sth->fetchrow()) {
            $maia_user_id = $row["maia_user_id"];
            delete_user($maia_user_id);
        }
        $sth->free();
    }


     /*
     * report_spam(): Confirm a mail item as false negative spam for one recipient.
     */
    function report_spam($user_id, $mail_id)
    {
        global $dbh;

        $expired_token = 'expired' . $user_id . $mail_id;

        // Mark the item as Confirmed Spam
        $update = "UPDATE maia_mail_recipients SET type = 'C', token=? " .
                  "WHERE recipient_id = ? AND mail_id = ?";
        $dbh->query($update, array($expired_token, $user_id, $mail_id));

        // Add the item to the Spam stats columns
        record_mail_stats($user_id, $mail_id, "fn");

    }

    /*
     * confirm_spam(): Confirm a mail item as spam for one recipient.
     */
    function confirm_spam($user_id, $mail_id)
    {
        global $dbh;

        $expired_token = 'expired' . $user_id . $mail_id;

        // Mark the item as Confirmed Spam
        $update = "UPDATE maia_mail_recipients SET type = 'C', token=? " .
                  "WHERE recipient_id = ? AND mail_id = ?";
        $dbh->query($update, array($expired_token, $user_id, $mail_id));

        // Add the item to the Spam stats columns
        record_mail_stats($user_id, $mail_id, "spam");
    }


    /*
     * confirm_ham(): Confirm a mail item as ham for one recipient.
     */
    function confirm_ham($user_id, $mail_id)
    {
    	global $dbh;

        $expired_token = 'expired' . $user_id . $mail_id;

        // Mark the item as Confirmed Ham
        $update = "UPDATE maia_mail_recipients SET type = 'G', token=? " .
                  "WHERE recipient_id = ? AND mail_id = ?";
        $dbh->query($update, array($expired_token, $user_id, $mail_id));

        // Add the item to the Ham stats columns
        record_mail_stats($user_id, $mail_id, "ham");
    }


    /*
     * rescue_item(): Rescues a mail item from quarantine and
     *                submits it for delivery.
     */
    function rescue_item($user_id, $mail_id)
    {
        global $dbh;

        $expired_token = 'expired' . $user_id . $mail_id;

        $select = "SELECT sender_email, contents, " .
                         "envelope_to, maia_mail_recipients.type " .
                  "FROM maia_mail, maia_mail_recipients " .
                  "WHERE maia_mail.id = maia_mail_recipients.mail_id " .
                  "AND maia_mail_recipients.recipient_id = ? " .
                  "AND maia_mail_recipients.mail_id = ?";
        $sth = $dbh->query($select, array($user_id, $mail_id));
        if ($row = $sth->fetchrow()) {
    	    $sender_email = $row["sender_email"];
    	    $body = $row["contents"];
    	    $type = $row["type"];

            if (extension_loaded('mcrypt')) {
       	        if (text_is_encrypted($body)) {
    	            $key = get_encryption_key();
    	            $body = decrypt_text($key, $body);
    	        }
            }

    	    if (is_a_domain_default_user($user_id)) {

    	    	// System default user (@.) or domain-class user (e.g. @domain)
    	    	$my_email_address = $row["envelope_to"];

    	    } else {

    	    	// Regular user (e.g. user@domain)
                $rlist = explode(" ", trim($row["envelope_to"]));
                $select = "SELECT email FROM users " .
                          "WHERE maia_user_id = ? " .
                          "AND email = ?";
                $my_email_address = "";
                foreach ($rlist as $rmail) {
                    $sth2 = $dbh->query($select, array($user_id, $rmail));
                    if ($row2 = $sth2->fetchrow()) {
                        $my_email_address = $row2["email"];
                        $sth2->free();
                        break;
                    }
                    $sth2->free();
                }

            }
            if (!empty($my_email_address)) {
                $smtp_result = smtp_send($sender_email, $my_email_address, $body);
                if ($succeeded = (strncmp($smtp_result, "2", 1) == 0)) {
            	    if ($type == "S") {
                        record_mail_stats($user_id, $mail_id, "fp");
                        $update = "UPDATE maia_mail_recipients SET type = 'G', token=? " .
                                  "WHERE recipient_id = ? AND mail_id = ?";
                        $dbh->query($update, array($expired_token, $user_id, $mail_id));
                        if (get_user_value($user_id, "auto_whitelist") == "Y") {
                            add_address_to_wb_list($user_id, $sender_email, "W");
                        }
                    } else {
                        delete_mail_reference($user_id, $mail_id);
            	    }
                } else {
                    trigger_error("rescue attempt failed!");
                }
            }
        }
        $sth->free();

        return $smtp_result;
    }


    /*
     * update_mail_stats(): Recalculate the statistics of the specified
     *                      type (suspected spam/ham), to ensure their
     *                      integrity after items of this type have been
     *                      deleted.
     */
    function update_mail_stats($user_id, $type)
    {
        global $dbh;

        if ($type == "suspected_spam") {
            $token = "S";
        } elseif ($type == "suspected_ham") {
            $token = "H";
        } else {
            $token = "";
        }

        if (!empty($token)) {

            $select = "SELECT MIN(received_date) AS mindate, " .
                             "MAX(received_date) AS maxdate, " .
                             "MIN(score) AS minscore, " .
                             "MAX(score) AS maxscore, " .
                             "SUM(score) AS totalscore, " .
                             "MIN(size) AS minsize, " .
                             "MAX(size) AS maxsize, " .
                             "SUM(size) AS totalsize, " .
                             "COUNT(id) AS items " .
                      "FROM maia_mail, maia_mail_recipients " .
                      "WHERE maia_mail.id = maia_mail_recipients.mail_id " .
                      "AND maia_mail_recipients.type = ? " .
                      "AND maia_mail_recipients.recipient_id = ?";
            $sth = $dbh->query($select, array($token, $user_id));
            if ($row = $sth->fetchrow()) {

            	$select = "SELECT user_id FROM maia_stats WHERE user_id = ?";
            	$sth2 = $dbh->query($select, array($user_id));

            	// User already has a stats record, update it.
            	if ($sth2->fetchrow()) {
                    $update = "UPDATE maia_stats SET oldest_" . $type . "_date = ?, " .
                                                    "newest_" . $type . "_date = ?, " .
                                                    "lowest_" . $type . "_score = ?, " .
                                                    "highest_" . $type . "_score = ?, " .
                                                    "total_" . $type . "_score = ?, " .
                                                    "smallest_" . $type . "_size = ?, " .
                                                    "largest_" . $type . "_size = ?, " .
                                                    "total_" . $type . "_size = ?, " .
                                                    "total_" . $type . "_items = ? " .
                              "WHERE user_id = ?";
                    $dbh->query($update, array($row["mindate"],
                                               $row["maxdate"],
                                               (isset($row["minscore"]) ? $row["minscore"] : 0),
                                               (isset($row["maxscore"]) ? $row["maxscore"] : 0),
                                               (isset($row["totalscore"]) ? $row["totalscore"] : 0),
                                               (isset($row["minsize"]) ? $row["minsize"] : 0),
                                               (isset($row["maxsize"]) ? $row["maxsize"] : 0),
                                               (isset($row["totalsize"]) ? $row["totalsize"] : 0),
                                               (isset($row["items"]) ? $row["items"] : 0),
                                               $user_id));

                // User doesn't have a stats record yet, create a new one for him.
                } else {
                    $insert = "INSERT INTO maia_stats (oldest_" . $type . "_date, " .
                                                      "newest_" . $type . "_date, " .
                                                      "lowest_" . $type . "_score, " .
                                                      "highest_" . $type . "_score, " .
                                                      "total_" . $type . "_score, " .
                                                      "smallest_" . $type . "_size, " .
                                                      "largest_" . $type . "_size, " .
                                                      "total_" . $type . "_size, " .
                                                      "total_" . $type . "_items, " .
                                                      "user_id) " .
                              "VALUES (?,?,?,?,?,?,?,?,?,?)";
                    $dbh->query($insert, array($row["mindate"],
                                               $row["maxdate"],
                                               (isset($row["minscore"]) ? $row["minscore"] : 0),
                                               (isset($row["maxscore"]) ? $row["maxscore"] : 0),
                                               (isset($row["totalscore"]) ? $row["totalscore"] : 0),
                                               (isset($row["minsize"]) ? $row["minsize"] : 0),
                                               (isset($row["maxsize"]) ? $row["maxsize"] : 0),
                                               (isset($row["totalsize"]) ? $row["totalsize"] : 0),
                                               (isset($row["items"]) ? $row["items"] : 0),
                                               $user_id));
                }
                $sth2->free();
            }
            $sth->free();
        }
    }


    /*
     * record_mail_stats(): Recalculate the specified user's stats record
     *                      for items of the specified (static) type, after
     *                      suspected [spam|ham] is confirmed.
     */
    function record_mail_stats($euid, $mail_id, $type)
    {
        global $dbh;

        $select = "SELECT received_date, size, score " .
                  "FROM maia_mail WHERE id = ?";
        $sth = $dbh->query($select, array($mail_id));
        if ($row = $sth->fetchrow()) {
            $mail_received_date = $row["received_date"];
            $mail_size = $row["size"];
            $mail_score = (isset($row["score"]) ? $row["score"] : 0);

            $select = "SELECT oldest_" . $type . "_date, " .
                             "newest_" . $type . "_date, " .
                             "lowest_" . $type . "_score, " .
                             "highest_" . $type . "_score, " .
                             "total_" . $type . "_score, " .
                             "smallest_" . $type . "_size, " .
                             "largest_" . $type . "_size, " .
                             "total_" . $type . "_size, " .
                             "total_" . $type . "_items " .
                      "FROM maia_stats WHERE user_id = ?";
            $sth2 = $dbh->query($select, array($euid));
            if ($row2 = $sth2->fetchrow()) {
                $oldest_date = $row2["oldest_" . $type . "_date"];
                $newest_date = $row2["newest_" . $type . "_date"];
                $lowest_score = $row2["lowest_" . $type . "_score"];
                $highest_score = $row2["highest_" . $type . "_score"];
                $total_score = $row2["total_" . $type . "_score"];
                $smallest_size = $row2["smallest_" . $type . "_size"];
                $largest_size = $row2["largest_" . $type . "_size"];
                $total_size = $row2["total_" . $type . "_size"];
                $total_items = $row2["total_" . $type . "_items"];

                if ($total_items == 0) {
                    $oldest_date = $mail_received_date;
                    $newest_date = $mail_received_date;
                    $lowest_score = $mail_score;
                    $highest_score = $mail_score;
                    $total_score = $mail_score;
                    $smallest_size = $mail_size;
                    $largest_size = $mail_size;
                    $total_size = $mail_size;
                    $total_items = 1;
                } else {
                    if ($oldest_date == NULL || $mail_received_date < $oldest_date) {
                        $oldest_date = $mail_received_date;
                    }
                    if ($mail_received_date > $newest_date) {
                        $newest_date = $mail_received_date;
                    }
                    if ($mail_score < $lowest_score) {
                        $lowest_score = $mail_score;
                    }
                    if ($mail_score > $highest_score) {
                        $highest_score = $mail_score;
                    }
                    $total_score += $mail_score;
                    if ($mail_size < $smallest_size) {
                        $smallest_size = $mail_size;
                    }
                    if ($mail_size > $largest_size) {
                        $largest_size = $mail_size;
                    }
                    $total_size += $mail_size;
                    $total_items++;
                }
                $update = "UPDATE maia_stats SET oldest_" . $type . "_date = ?, " .
                                                "newest_" . $type . "_date = ?, " .
                                                "lowest_" . $type . "_score = ?, " .
                                                "highest_" . $type . "_score = ?, " .
                                                "total_" . $type . "_score = ?, " .
                                                "smallest_" . $type . "_size = ?, " .
                                                "largest_" . $type . "_size = ?, " .
                                                "total_" . $type . "_size = ?, " .
                                                "total_" . $type . "_items = ? " .
                          "WHERE user_id = ?";
                $dbh->query($update, array($oldest_date,
                                           $newest_date,
                                           $lowest_score,
                                           $highest_score,
                                           $total_score,
                                           $smallest_size,
                                           $largest_size,
                                           $total_size,
                                           $total_items,
                                           $euid));

            } else {
                $oldest_date = $mail_received_date;
                $newest_date = $mail_received_date;
                $lowest_score = $mail_score;
                $highest_score = $mail_score;
                $total_score = $mail_score;
                $smallest_size = $mail_size;
                $largest_size = $mail_size;
                $total_size = $mail_size;
                $insert = "INSERT INTO maia_stats (oldest_" . $type . "_date, " .
                                                  "newest_" . $type . "_date, " .
                                                  "lowest_" . $type . "_score, " .
                                                  "highest_" . $type . "_score, " .
                                                  "total_" . $type . "_score, " .
                                                  "smallest_" . $type . "_size, " .
                                                  "largest_" . $type . "_size, " .
                                                  "total_" . $type . "_size, " .
                                                  "total_" . $type . "_items, " .
                                                  "user_id) " .
                          "VALUES (?,?,?,?,?,?,?,?,1,?)";
                $dbh->query($insert, array($oldest_date,
                                           $newest_date,
                                           $lowest_score,
                                           $highest_score,
                                           $total_score,
                                           $smallest_size,
                                           $largest_size,
                                           $total_size,
                                           $euid));
            }
            $sth2->free();
        }
        $sth->free();
    }
    
    /*
     *  function count_cache_items($euid) returns an 
     *      array of item counts for a given user
     */
    function count_cache_items($euid) { 
      if ($euid == 0) {
        return array(); // at some time, we may want to return the total count, 
                        // but the current desert_sand theme calls this during
                        // login.php, and that could be an expensive query.
      }
      global $dbh;
      $select = "SELECT type, 
                        COUNT(mail_id) AS qcount,
                        MAX(mail_id) AS maxid 
                 FROM maia_mail_recipients 
                 WHERE recipient_id = ? 
                 GROUP BY type";
      $sth = $dbh->query($select, $euid);
      $results = array();
      while ($row = $sth->fetchRow()) {          
          $results[$row['type']] = array('count' => $row['qcount'],
                                         'max'   => $row['maxid']);
      }
      $sth->free();
      return $results;
    }

    /*
     * get_wb_status(): Returns 'W' if a given e-mail address appears in a given user's
     *                  whitelist, 'B' if it appears in the user's blacklist, or '' otherwise.
     */
    function get_wb_status($user_id, $addr_id)
    {
       global $dbh;

       $wb = '';
       $select = "SELECT wb " .
                 "FROM wblist " .
                 "WHERE rid = ? " .
                 "AND sid = ?";
       $sth = $dbh->query($select, array($user_id, $addr_id));
       if ($row = $sth->fetchRow()) {
          $wb = $row["wb"];
       }
       $sth->free();

       return $wb;
    }


    /*
     * set_wb_status(): Changes the status ('W' or 'B') of an entry in a given user's
     *                  whitelist/blacklist.
     */
    function set_wb_status($user_id, $addr_id, $wb)
    {
       global $dbh;

       $update = "UPDATE wblist SET wb = ? WHERE rid = ? AND sid = ?";
       $dbh->query($update, array($wb, $user_id, $addr_id));
    }


    /*
     * add_wb_entry(): Adds a new entry to a user's whitelist/blacklist.
     */
    function add_wb_entry($user_id, $addr_id, $wb)
    {
       global $dbh;

       $insert = "INSERT INTO wblist (rid, sid, wb) VALUES (?, ?, ?)";
       $dbh->query($insert, array($user_id, $addr_id, $wb));
    }


    /*
     * delete_wb_entry(): Removes an entry from a user's whitelist/blacklist.
     */
    function delete_wb_entry($user_id, $addr_id)
    {
       global $dbh;

       $delete = "DELETE FROM wblist WHERE rid = ? AND sid = ?";
       $dbh->query($delete, array($user_id, $addr_id));

       // If there are no other references to this address,
       // remove it from the mailaddr table as well.
       $select = "SELECT wb FROM wblist WHERE sid = ?";
       $sth2 = $dbh->query($select, array($addr_id));
       if (!$sth2->fetchRow()) {
           $delete = "DELETE FROM mailaddr WHERE id = ?";
           $dbh->query($delete, array($addr_id));
       }
       $sth2->free();
    }


    /*
     * delete_user_wb_entries(): Removes all entries from a user's whitelist/blacklist.
     */
    function delete_user_wb_entries($user_id)
    {
       global $dbh;

       $select = "SELECT sid " .
                 "FROM wblist " .
                 "WHERE rid = ?";
       $sth = $dbh->query($select, array($user_id));
       while ($row = $sth->fetchRow()) {
           delete_wb_entry($user_id, $row["sid"]);
       }
       $sth->free();
    }


    /*
     * add_address(): Adds a new e-mail address to the mailaddr table.
     *                Note: Assumes that address is either in "user@domain" form
     *                      or "@domain" form.  i.e. call fix_address() first.
     */
    function add_address($email)
    {
       global $dbh;

       $addr_id = 0;
       $priority = get_email_address_priority($email);
       $insert = "INSERT INTO mailaddr (priority, email) VALUES (?, ?)";
       $dbh->query($insert, array($priority, $email));
       $select = "SELECT id FROM mailaddr WHERE email = ?";
       $sth = $dbh->query($select, array($email));
       if ($row = $sth->fetchRow()) {
           $addr_id = $row["id"];
       }
       $sth->free();

       return $addr_id;
    }


    /*
     * get_address_id(): Looks up the ID number of an e-mail address in
     *                   the mailaddr table.
     */
    function get_address_id($addr)
    {
       global $dbh;

       $select = "SELECT id FROM mailaddr WHERE email = ?";

       $addr_id = 0;
       $sth = $dbh->query($select, array($addr));
       if ($row = $sth->fetchRow()) {
          $addr_id = $row["id"];
       }
       $sth->free();

       return $addr_id;
    }


    /*
     * add_address_to_wb_list(): Adds an e-mail address to a user's
     *                           whitelist/blacklist, if necessary.
     *     returns message code, use to match with locale strings.
     */
    function add_address_to_wb_list($user_id, $addr, $wb)
    {
        global $dbh;

        $addr = fix_address($addr);
        $addr_id = get_address_id($addr);
        if ($addr_id == 0) {
            $addr_id = add_address($addr);
            add_wb_entry($user_id, $addr_id, $wb);
            return 'text_wb_address_added';
        } else {
            $wb_stat = get_wb_status($user_id, $addr_id);
            if ($wb_stat == '') {
                add_wb_entry($user_id, $addr_id, $wb);
                return 'text_wb_address_added';
            } else {
                if ($wb_stat != $wb) {
                    set_wb_status($user_id, $addr_id, $wb);                    
                }
                return 'text_wb_address_changed';
            }
        }
    }


    /*
     * count_total_mail(): Returns the total number of mail items received
     *                     by the specified user, or by all users if $user_id == 0.
     */
    function count_total_mail($user_id)
    {
    	global $dbh;

        $select = "SELECT enable_false_negative_management, " .
                         "enable_virus_scanning, enable_spam_filtering, " .
                         "enable_bad_header_checking, enable_banned_files_checking " .
                  "FROM maia_config WHERE id = 0";
        $sth = $dbh->query($select);
        if ($row = $sth->fetchrow()) {
            $enable_false_negative_management = ($row["enable_false_negative_management"] == 'Y');
            $enable_virus_scanning = ($row["enable_virus_scanning"] == 'Y');
            $enable_spam_filtering = ($row["enable_spam_filtering"] == 'Y');
            $enable_bad_header_checking = ($row["enable_bad_header_checking"] == 'Y');
            $enable_banned_files_checking = ($row["enable_banned_files_checking"] == 'Y');
        }
        $sth->free();

    	if ($user_id > 0) {
    	    $prefix = "SELECT (";
    	    $suffix = " WHERE user_id = ?";
    	} else {
    	    $prefix = "SELECT SUM(";
    	    $suffix = "";
    	}
    	$select = $prefix . "total_ham_items + " .
    	         	    "total_wl_items + " .
    	                    "total_bl_items";
    	if ($enable_spam_filtering) {
    	    $select .= " + total_fp_items" .
    	               " + total_suspected_spam_items" .
    	               " + total_spam_items";
    	    if ($enable_false_negative_management) {
    	        $select .= " + total_suspected_ham_items" .
    	                   " + total_fn_items";
    	    }
        }
        if ($enable_virus_scanning) {
    	    $select .= " + total_virus_items";
    	}
    	if ($enable_bad_header_checking) {
    	    $select .= " + total_bad_header_items";
    	}
    	if ($enable_banned_files_checking) {
    	    $select .= " + total_banned_file_items";
    	}
    	$select .= ") AS count FROM maia_stats" . $suffix;

        if ($user_id > 0) {
    	    $sth = $dbh->query($select, array($user_id));
    	} else {
    	    $sth = $dbh->query($select);
    	}

    	$count = 0;
        if ($row = $sth->fetchrow()) {
            $count = $row["count"];
        }
        $sth->free();

        return $count;
    }


    /*
     * count_item_days(): Returns the number of days between the oldest and
     *                    newest mail item of the specified type received by
     *                    the specified user, or all users if $user_id == 0.
     *                    $type can be one of:
     *
     *                        "suspected_ham"  : Suspected Ham
     *                        "ham"            : Confirmed Ham
     *                        "suspected_spam" : Suspected Spam
     *                        "spam"           : Confirmed Spam
     *                        "wl"             : Whitelisted Sender
     *                        "bl"             : Blacklisted Sender
     *                        "fp"             : False Positive
     *                        "fn"             : False Negative
     *                        "virus"          : Virus/Malware
     *                        "banned_file"    : Banned File Attachment
     *                        "bad_header"     : Invalid Mail Header
     */
    function count_item_days($user_id, $type)
    {
        global $dbh;

        if ($user_id > 0) {
            $select = "SELECT oldest_" . $type . "_date AS mindate FROM maia_stats WHERE user_id = ?";
            $sth = $dbh->query($select, array($user_id));
        } else {
            $select = "SELECT MIN(oldest_" . $type . "_date) AS mindate FROM maia_stats";
            $sth = $dbh->query($select);
        }

        $days = 0.0;
        if ($row = $sth->fetchrow()) {
            sscanf($row["mindate"], "%d-%d-%d %d:%d:%d", $year, $mon, $day, $hour, $min, $sec);
            $tstamp = mktime($hour, $min, $sec, $mon, $day, $year);
            $days = ((time() - $tstamp) / (24 * 60 * 60));
        }
        $sth->free();

        return $days;
    }


    /*
     * lowest_item_score(): Returns the lowest SpamAssassin score recorded
     *                      for a mail item of the specified type for the
     *                      specified user, or all users if $user_id == 0.
     *                      $type can be one of:
     *
     *                          "suspected_ham"  : Suspected Ham
     *                          "ham"            : Confirmed Ham
     *                          "suspected_spam" : Suspected Spam
     *                          "spam"           : Confirmed Spam
     *                          "fp"             : False Positive
     *                          "fn"             : False Negative
     */
    function lowest_item_score($user_id, $type)
    {
    	global $dbh;

    	if ($user_id > 0) {
    	    $select = "SELECT lowest_" . $type . "_score AS score FROM maia_stats " .
    	              "WHERE user_id = ? AND lowest_" . $type . "_score > -999";
    	    $sth = $dbh->query($select, array($user_id));
    	} else {
    	    $select = "SELECT MIN(lowest_" . $type . "_score) AS score FROM maia_stats " .
    	              "WHERE lowest_" . $type . "_score > -999";
            $sth = $dbh->query($select);
    	}

    	$score = 0;
        if ($row = $sth->fetchrow()) {
            $score = $row["score"];
        }
        $sth->free();

        return $score;
    }


    /*
     * highest_item_score(): Returns the highest SpamAssassin score recorded
     *                       for a mail item of the specified type for the
     *                       specified user, or all users if $user_id == 0.
     *                       $type can be one of:
     *
     *                           "suspected_ham"  : Suspected Ham
     *                           "ham"            : Confirmed Ham
     *                           "suspected_spam" : Suspected Spam
     *                           "spam"           : Confirmed Spam
     *                           "fp"             : False Positive
     *                           "fn"             : False Negative
     */
    function highest_item_score($user_id, $type)
    {
    	global $dbh;

    	if ($user_id > 0) {
    	    $select = "SELECT highest_" . $type . "_score AS score FROM maia_stats " .
    	              "WHERE user_id = ? AND highest_" . $type . "_score < 999";
    	    $sth = $dbh->query($select, array($user_id));
    	} else {
    	    $select = "SELECT MAX(highest_" . $type . "_score) AS score FROM maia_stats " .
    	              "WHERE highest_" . $type . "_score < 999";
            $sth = $dbh->query($select);
    	}

    	$score = 0;
        if ($row = $sth->fetchrow()) {
            $score = $row["score"];
        }
        $sth->free();

        return $score;
    }


    /*
     * total_item_score(): Returns the sum of all SpamAssassin scores
     *                     recorded for a mail item of the specified
     *                     type for the specified user, or all users
     *                     if $user_id == 0.  $type can be one of:
     *
     *                         "suspected_ham"  : Suspected Ham
     *                         "ham"            : Confirmed Ham
     *                         "suspected_spam" : Suspected Spam
     *                         "spam"           : Confirmed Spam
     *                         "fp"             : False Positive
     *                         "fn"             : False Negative
     */
    function total_item_score($user_id, $type)
    {
    	global $dbh;

    	if ($user_id > 0) {
    	    $select = "SELECT total_" . $type . "_score AS score FROM maia_stats WHERE user_id = ?";
    	    $sth = $dbh->query($select, array($user_id));
    	} else {
    	    $select = "SELECT SUM(total_" . $type . "_score) AS score FROM maia_stats";
            $sth = $dbh->query($select);
    	}

    	$score = 0;
        if ($row = $sth->fetchrow()) {
            $score = $row["score"];
        }
        $sth->free();

        return $score;
    }


    /*
     * smallest_item_size(): Returns the size of the smallest mail item
     *                       of the specified type received by the specified
     *                       user, or all users if $user_id == 0.  $type can
     *                       be one of:
     *
     *                           "suspected_ham"  : Suspected Ham
     *                           "ham"            : Confirmed Ham
     *                           "suspected_spam" : Suspected Spam
     *                           "spam"           : Confirmed Spam
     *                           "wl"             : Whitelisted Sender
     *                           "bl"             : Blacklisted Sender
     *                           "fp"             : False Positive
     *                           "fn"             : False Negative
     *                           "virus"          : Virus/Malware
     *                           "banned_file"    : Banned File Attachment
     *                           "bad_header"     : Invalid Mail Header
     */
    function smallest_item_size($user_id, $type)
    {
    	global $dbh;

    	if ($user_id > 0) {
    	    $select = "SELECT smallest_" . $type . "_size AS size FROM maia_stats WHERE user_id = ? AND smallest_" . $type . "_size > 0";
    	    $sth = $dbh->query($select, array($user_id));
    	} else {
    	    $select = "SELECT MIN(smallest_" . $type . "_size) AS size FROM maia_stats WHERE smallest_" . $type . "_size > 0";
            $sth = $dbh->query($select);
    	}

    	$size = 0;
        if ($row = $sth->fetchrow()) {
            $size = $row["size"];
        }
        $sth->free();

        return $size;
    }


    /*
     * largest_item_size(): Returns the size of the largest mail item
     *                      of the specified type received by the specified
     *                      user, or all users if $user_id == 0.  $type can
     *                      be one of:
     *
     *                          "suspected_ham"  : Suspected Ham
     *                          "ham"            : Confirmed Ham
     *                          "suspected_spam" : Suspected Spam
     *                          "spam"           : Confirmed Spam
     *                          "wl"             : Whitelisted Sender
     *                          "bl"             : Blacklisted Sender
     *                          "fp"             : False Positive
     *                          "fn"             : False Negative
     *                          "virus"          : Virus/Malware
     *                          "banned_file"    : Banned File Attachment
     *                          "bad_header"     : Invalid Mail Header
     */
    function largest_item_size($user_id, $type)
    {
    	global $dbh;

    	if ($user_id > 0) {
    	    $select = "SELECT largest_" . $type . "_size AS size FROM maia_stats WHERE user_id = ?";
    	    $sth = $dbh->query($select, array($user_id));
    	} else {
    	    $select = "SELECT MAX(largest_" . $type . "_size) AS size FROM maia_stats";
            $sth = $dbh->query($select);
    	}

    	$size = 0;
        if ($row = $sth->fetchrow()) {
            $size = $row["size"];
        }
        $sth->free();

        return $size;
    }


    /*
     * total_item_size(): Returns the sum of the sizes of all mail items
     *                    of the specified type received by the specified
     *                    user, or all users if $user_id == 0.  $type can
     *                    be one of:
     *
     *                        "suspected_ham"  : Suspected Ham
     *                        "ham"            : Confirmed Ham
     *                        "suspected_spam" : Suspected Spam
     *                        "spam"           : Confirmed Spam
     *                        "wl"             : Whitelisted Sender
     *                        "bl"             : Blacklisted Sender
     *                        "fp"             : False Positive
     *                        "fn"             : False Negative
     *                        "virus"          : Virus/Malware
     *                        "banned_file"    : Banned File Attachment
     *                        "bad_header"     : Invalid Mail Header
     */
    function total_item_size($user_id, $type)
    {
    	global $dbh;

    	if ($user_id > 0) {
    	    $select = "SELECT total_" . $type . "_size AS size FROM maia_stats WHERE user_id = ?";
    	    $sth = $dbh->query($select, array($user_id));
    	} else {
    	    $select = "SELECT SUM(total_" . $type . "_size) AS size FROM maia_stats";
            $sth = $dbh->query($select);
    	}

    	$size = 0;
        if ($row = $sth->fetchrow()) {
            $size = $row["size"];
        }
        $sth->free();

        return $size;
    }


    /*
     * count_items(): Returns the total number of mail items of the specified
     *                type received by the specified user, or all users if
     *                $user_id == 0.  $type can be one of:
     *
     *                    "suspected_ham"  : Suspected Ham
     *                    "ham"            : Confirmed Ham
     *                    "suspected_spam" : Suspected Spam
     *                    "spam"           : Confirmed Spam
     *                    "wl"             : Whitelisted Sender
     *                    "bl"             : Blacklisted Sender
     *                    "fp"             : False Positive
     *                    "fn"             : False Negative
     *                    "virus"          : Virus/Malware
     *                    "banned_file"    : Banned File Attachment
     *                    "bad_header"     : Invalid Mail Header
     */
    function count_items($user_id, $type)
    {
    	global $dbh;

    	if ($user_id > 0) {
    	    $select = "SELECT total_" . $type . "_items AS count FROM maia_stats WHERE user_id = ?";
    	    $sth = $dbh->query($select, array($user_id));
    	} else {
    	    $select = "SELECT SUM(total_" . $type . "_items) AS count FROM maia_stats";
            $sth = $dbh->query($select);
    	}

    	$count = 0;
        if ($row = $sth->fetchrow()) {
            $count = $row["count"];
        }
        $sth->free();

        return $count;
    }


    /*
     * calculate_ppv(): Calculates the Positive Predictive Value (PPV), which is
      *                  the percentage of Suspected Spam that ultimately becomes
      *                  Confirmed Spam.
      */
    function calculate_ppv($spam, $fp)
    {
        $total = $spam + $fp;
          if ($total > 0) {
              return ($spam / $total);
          } else {
              return 0;
          }
      }
  
  
      /*
      * calculate_npv(): Calculates the Negative Predictive Value (NPV), which is
      *                  the percentage of Suspected Ham that ultimately becomes
      *                  Confirmed Ham.
      */
      function calculate_npv($ham, $fn)
      {
          $total = $ham + $fn;
          if ($total > 0) {
              return ($ham / $total);
          } else {
              return 0;
          }
      }
  
  
      /*
      * calculate_sensitivity(): Calculates the statistical sensitivity, which is
      *                          the percentage of Confirmed Spam that was properly
      *                          diagnosed as Suspected Spam.
      */
      function calculate_sensitivity($spam, $fn)
      {
          $total = $spam + $fn;
          if ($total > 0) {
              return ($spam / $total);
          } else {
              return 0;
          }
      }
  
  
      /*
      * calculate_specificity(): Calculates the statistical specificity, which is
      *                          the percentage of Confirmed Ham that was properly
      *                          diagnosed as Suspected Ham.
     */
     function calculate_specificity($ham, $fp)
    {
        $total = $ham + $fp;
        if ($total > 0) {
            return ($ham / $total);
        } else {
            return 0;
        }
    }


    /*
    * calculate_efficiency(): Calculates the statistical efficiency, which is the
    *                         percentage of the time SpamAssassin diagnosed the
    *                         mail's status correctly.
    */
    function calculate_efficiency($spam, $ham, $fp, $fn)
    {
        $total = $spam + $ham + $fp + $fn;
        if ($total > 0) {
            return (($spam + $ham) / $total);
        } else {
            return 0;
        }
    }


    /*
    * get_filter_stats():
    */
    function get_filter_stats($user_id)
    {
        global $dbh;

        $enable_false_negative_management = (get_config_value("enable_false_negative_management") == 'Y');
        if ($user_id > 0) {
            $select = "SELECT total_ham_items AS total_ham, " .
                             "total_spam_items AS total_spam, " .
                             "total_fp_items AS total_fp, " .
                             "total_fn_items AS total_fn " .
                      "FROM maia_stats WHERE user_id = ?";
            $sth = $dbh->query($select, array($user_id));
        } else {
            $select = "SELECT SUM(total_ham_items) AS total_ham, " .
                             "SUM(total_spam_items) AS total_spam, " .
                             "SUM(total_fp_items) AS total_fp, " .
                             "SUM(total_fn_items) AS total_fn " .
                      "FROM maia_stats";
            $sth = $dbh->query($select);
        }
        $fp_pct = 0;
        $fn_pct = 0;
        $ppv_pct = 0;
        $npv_pct = 0;
        $sensitivity_pct = 0;
        $specificity_pct = 0;
        $efficiency_pct = 0;
        if ($row = $sth->fetchrow()) {
            $ham = $row["total_ham"];
            $spam = $row["total_spam"];
            $fp = $row["total_fp"];
            $fn = $row["total_fn"];
            $ppv_pct = calculate_ppv($spam, $fp);
            $specificity_pct = calculate_specificity($ham, $fp);
            if ($enable_false_negative_management) {
                  $npv_pct = calculate_npv($ham, $fn);
                  $sensitivity_pct = calculate_sensitivity($spam, $fn);
                  $efficiency_pct = calculate_efficiency($spam, $ham, $fp, $fn);
                  if ($ham + $spam + $fp + $fn > 0) {
                      $fp_pct = $fp / ($ham + $spam + $fp + $fn);
                      $fn_pct = 1 - $efficiency_pct - $fp_pct;
                  } else {
                      $fp_pct = 0;
                      $fn_pct = 0;
                  }
            } else {
                  $npv = 1;
                  $sensitivity_pct = 1;
                  $efficiency_pct = calculate_efficiency($spam, $ham, $fp, 0);
                  if ($ham + $spam + $fp > 0) {
                      $fp_pct = 1 - $efficiency_pct;
                  } else {
                      $fp_pct = 0;
                  }
            	  $fn_pct = 0;
            }
        }
        $sth->free();

    	return array($ppv_pct, $npv_pct, $sensitivity_pct, $specificity_pct,
  	                      $efficiency_pct, $fp_pct, $fn_pct);
    }


    /*
     * get_display_language(): Determine which language to use for
     *                         HTML display purposes.  User-specified
     *                         choices are preferred, but the default
     *                         language (specified in config.php) will
     *                         be used if no other language is
     *                         specified.
     */
    function get_display_language($user_id)
    {
    	global $dbh;
    	global $default_display_language;

        if (isset($_SESSION["display_language"])) {
            $display_language = trim($_SESSION["display_language"]);
        } else {
            $select = "SELECT language FROM maia_users WHERE id = ?";
            $sth = $dbh->query($select, array($user_id));
            if ($row = $sth->fetchrow()) {
                $display_language = strtolower($row["language"]);
                $_SESSION["display_language"] = $display_language;
            } else {
                $display_language = $default_display_language;
            }
            $sth->free();
        }

        return $display_language;
    }


    /*
     * get_html_charset(): Determine which character set to use for
     *                     HTML display purposes.  User-specified
     *                     choices are preferred, but the default
     *                     character set (specified in core.php) will
     *                     be used if no other character set is specified.
     */
    function get_html_charset($user_id)
    {
    	global $dbh;
    	global $default_charset;

        if (isset($_SESSION["html_charset"])) {
            $html_charset = trim($_SESSION["html_charset"]);
        } else {
            $select = "SELECT charset FROM maia_users WHERE id = ?";
            $sth = $dbh->query($select, array($user_id));
            if ($row = $sth->fetchrow()) {
                $html_charset = strtolower($row["charset"]);
                $_SESSION["html_charset"] = $html_charset;
            } else {
                $html_charset = $default_charset;
            }
            $sth->free();
        }

        return $html_charset;
    }


    /*
     * get_config_value(): Returns the value corresponding to a single
     *                     key from the system configuration table.  This
     *                     is a convenient way to read one configuration
     *                     setting, but if multiple configuration settings
     *                     are required, it becomes more efficient to
     *                     use a manual query, rather than calling this
     *                     function repeatedly.
     */
    function get_config_value($key)
    {
        global $dbh;

        $select = "SELECT " . $key . " FROM maia_config WHERE id = 0";
        $sth = $dbh->query($select);

        if ($row = $sth->fetchrow()) {
            $value = $row[$key];
        }

        return $value;
    }


    /*
     * get_user_value(): Returns a value corresponding to a single
     *                   key from a user's personal settings.  This
     *                   is a convenient way to read one configuration
     *                   setting, but if multiple configuration settings
     *                   are required, it becomes more efficient to use
     *                   a manual query, rather than calling this
     *                   function repeatedly.
     */
    function get_user_value($user_id, $key)
    {
        global $dbh;
       
        $select = "SELECT " . $key . " FROM maia_users WHERE id = ?";
        $sth = $dbh->query($select, array($user_id));

        if ($row = $sth->fetchrow()) {
            $value = $row[$key];
        }

        return $value;
    }
    
     /*
     * get_maia_user_row(): Returns a hash corresponding to a user's personal settings.  
     *                      Call this to get all values in a user's config.
     */
    function get_maia_user_row($user_id)
    {
        global $dbh;
       
        $select = "SELECT * FROM maia_users WHERE id = ? LIMIT 1";
        $sth = $dbh->query($select, array($user_id));

        if ($row = $sth->fetchrow()) {
            $value = $row;
        }

        return $value;
    }
     /*
      * generate_random_password(): Generates an 8-character alphanumeric
      *                             password, and returns both the password
      *                             and its corresponding MD5 hash.
      */
      function generate_random_password()
      {
          $password_length = 8;
  
          for ($i = "A"; $i < "Z"; $i++) {
              $chars[] = $i;
          }
          for ($i = "a"; $i < "z"; $i++) {
              $chars[] = $i;
          }
          for ($i = "0"; $i < "9"; $i++) {
              $chars[] = $i;
          }
  
          $password = "";
          for ($i = 0; $i < $password_length; $i++) {
              $password .= $chars[mt_rand(0, count($chars)-1)];
          }
          $digest = md5($password);
  
          return array($password, $digest);
      }
      
      function get_chart_colors() {
        global $dbh;
        
        $select = "SELECT chart_suspected_ham_colour AS sh, chart_ham_colour AS h, chart_wl_colour as wl, " .
                          "chart_bl_colour as bl, chart_suspected_spam_colour as ss, chart_spam_colour as s, " .
                          "chart_fp_colour as fp, chart_fn_colour as fn, chart_virus_colour as v " .
                  "FROM maia_config WHERE id=0";
        
        $sth = $dbh->query($select);

        if ($row = $sth->fetchrow()) {
            $value = $row;
        }          
        $sth->free();
        
        $value["bf"] = "#C9BBFF";
        $value["bh"] = "#FFCC79";
        $value["os"] = "#FF8080";
        
        return $value;
      
      
      }
?>
